# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Server subcube of cubicweb : defines objects used only on the server
(repository) side

The server module contains functions to initialize a new repository.
"""
from contextlib import contextmanager

from logilab import database
from logilab.common.modutils import LazyObject
from logilab.common.registry import yes
from logilab.common.textutils import splitstrip
from yams import BASE_GROUPS

from cubicweb.appobject import AppObject


# server-side services #########################################################


class Service(AppObject):
    """Base class for services.

    A service is a selectable object that performs an action server-side.
    Use :class:`cubicweb.dbapi.Connection.call_service` to call them from
    the web-side.

    When inheriting this class, do not forget to define at least the __regid__
    attribute (and probably __select__ too).
    """

    __registry__ = "services"
    __select__ = yes()

    def call(self, **kwargs):
        raise NotImplementedError


# server-side debugging ########################################################

# server debugging flags. They may be combined using binary operators.

#: no debug information
DBG_NONE = 0  #: no debug information
#: rql execution information
DBG_RQL = 1
#: executed sql
DBG_SQL = 2
#: repository events
DBG_REPO = 4
#: hooks
DBG_HOOKS = 16
#: operations
DBG_OPS = 32
#: security
DBG_SEC = 64
#: more verbosity
DBG_MORE = 128
#: all level enabled
DBG_ALL = DBG_RQL + DBG_SQL + DBG_REPO + DBG_HOOKS + DBG_OPS + DBG_SEC + DBG_MORE

_SECURITY_ITEMS = []
_SECURITY_CAPS = ["read", "add", "update", "delete", "transition"]

#: current debug mode
DEBUG = 0


@contextmanager
def tunesecurity(items=(), capabilities=()):
    """Context manager to use in conjunction with DBG_SEC.

    This allows some tuning of:
    * the monitored capabilities ('read', 'add', ....)
    * the object being checked by the security checkers

    When no item is given, all of them will be watched.
    By default all capabilities are monitored, unless specified.

    Example use::

      from cubicweb.server import debugged, DBG_SEC, tunesecurity
      with debugged(DBG_SEC):
          with tunesecurity(items=('Elephant', 'trumps'),
                            capabilities=('update', 'delete')):
              babar.cw_set(trumps=celeste)
              flore.cw_delete()

      ==>

      check_perm: 'update' 'relation Elephant.trumps.Elephant'
       [(ERQLExpression(Any X WHERE U has_update_permission X, X eid %(x)s, U eid %(u)s),
       {'eid': 2167}, True)]
      check_perm: 'delete' 'Elephant'
       [(ERQLExpression(Any X WHERE U has_delete_permission X, X eid %(x)s, U eid %(u)s),
       {'eid': 2168}, True)]

    """
    olditems = _SECURITY_ITEMS[:]
    _SECURITY_ITEMS.extend(list(items))
    oldactions = _SECURITY_CAPS[:]
    _SECURITY_CAPS[:] = capabilities
    yield
    _SECURITY_ITEMS[:] = olditems
    _SECURITY_CAPS[:] = oldactions


def set_debug(debugmode):
    """change the repository debugging mode"""
    global DEBUG
    if not debugmode:
        DEBUG = 0
        return
    if isinstance(debugmode, str):
        for mode in splitstrip(debugmode, sep="|"):
            DEBUG |= globals()[mode]
    else:
        DEBUG |= debugmode


class debugged:
    """Context manager and decorator to help debug the repository.

    It can be used either as a context manager:

    >>> with debugged('DBG_RQL | DBG_REPO'):
    ...     # some code in which you want to debug repository activity,
    ...     # seing information about RQL being executed an repository events.

    or as a function decorator:

    >>> @debugged('DBG_RQL | DBG_REPO')
    ... def some_function():
    ...     # some code in which you want to debug repository activity,
    ...     # seing information about RQL being executed an repository events

    The debug mode will be reset to its original value when leaving the "with"
    block or the decorated function.
    """

    def __init__(self, debugmode):
        self.debugmode = debugmode
        self._clevel = None

    def __enter__(self):
        """enter with block"""
        self._clevel = DEBUG
        set_debug(self.debugmode)

    def __exit__(self, exctype, exc, traceback):
        """leave with block"""
        set_debug(self._clevel)
        return traceback is None

    def __call__(self, func):
        """decorate function"""

        def wrapped(*args, **kwargs):
            set_debug(self.debugmode)
            try:
                return func(*args, **kwargs)
            finally:
                set_debug(self._clevel)

        return wrapped


# database initialization ######################################################


def create_user(session, login, pwd, *groups):
    # monkey patch this method if you want to customize admin/anon creation
    # (that maybe necessary if you change CWUser's schema)
    user = session.create_entity("CWUser", login=login, upassword=pwd)
    for group in groups:
        session.execute(
            "SET U in_group G WHERE U eid %(u)s, G name %(group)s",
            {"u": user.eid, "group": group},
        )
    return user


def init_repository(config, interactive=True, drop=False, vreg=None, init_config=None):
    """Initialise a repository database by creating tables and filling them
    with the minimal set of entities (ie at least the schema, base groups and
    a initial user)
    """
    from cubicweb.repoapi import get_repository, connect
    from cubicweb.server.repository import Repository
    from cubicweb.server.utils import manager_userpasswd
    from cubicweb.server.sqlutils import sqlexec, sqlschema, sql_drop_all_user_tables
    from cubicweb.server.sqlutils import (
        _SQL_DROP_ALL_USER_TABLES_FILTER_FUNCTION as drop_filter,
    )

    # configuration to avoid db schema loading and user'state checking
    # on connection
    config.creating = True
    config.consider_user_state = False
    config.cubicweb_appobject_path = set(("hooks", "entities"))
    config.cube_appobject_path = set(("hooks", "entities"))
    # only enable the system source at initialization time
    repo = Repository(config, vreg=vreg)
    repo.bootstrap()
    if init_config is not None:
        # further config initialization once it has been bootstrapped
        init_config(config)
    schema = repo.schema
    sourcescfg = config.read_sources_file()
    source = sourcescfg["system"]
    driver = source["db-driver"]
    with repo.internal_cnx() as cnx:
        sqlcnx = cnx.cnxset.cnx
        sqlcursor = cnx.cnxset.cu
        execute = sqlcursor.execute
        if drop:
            helper = database.get_db_helper(driver)
            dropsql = sql_drop_all_user_tables(helper, sqlcursor)
            # We may fail dropping some tables because of table dependencies, in a first pass.
            # So, we try a second drop sequence to drop remaining tables if needed.
            # Note that 2 passes is an arbitrary choice as it seems enough for our usecases
            # (looping may induce infinite recursion when user have no rights for example).
            # Here we try to keep code simple and backend independent. That's why we don't try to
            # distinguish remaining tables (missing privileges, dependencies, ...).
            failed = sqlexec(
                dropsql, execute, cnx=sqlcnx, pbtitle="-> dropping tables (first pass)"
            )
            if failed:
                failed = sqlexec(
                    failed,
                    execute,
                    cnx=sqlcnx,
                    pbtitle="-> dropping tables (second pass)",
                )
                remainings = list(filter(drop_filter, helper.list_tables(sqlcursor)))
                assert not remainings, f"Remaining tables: {', '.join(remainings)}"
        handler = config.migration_handler(
            schema, interactive=False, repo=repo, cnx=cnx
        )
        # install additional driver specific sql files
        handler.cmd_install_custom_sql_scripts()
        for cube in reversed(config.cubes()):
            handler.cmd_install_custom_sql_scripts(cube)
        _title = "-> creating tables "
        print(_title, end=" ")
        # schema entities and relations tables
        # can't skip entities table even if system source doesn't support them,
        # they are used sometimes by generated sql. Keeping them empty is much
        # simpler than fixing this...
        schemasql = sqlschema(schema, driver)
        failed = sqlexec(schemasql, execute, pbtitle=_title)
        if failed:
            print("The following SQL statements failed. You should check your schema.")
            print(failed)
            raise Exception(
                "execution of the sql schema failed, you should check your schema"
            )
        sqlcnx.commit()
    with repo.internal_cnx() as cnx:
        # insert entity representing the system source
        ssource = cnx.create_entity("CWSource", type="native", name="system")
        repo.system_source.eid = ssource.eid
        cnx.execute("SET X cw_source X WHERE X eid %(x)s", {"x": ssource.eid})
        # insert base groups and default admin
        print("-> inserting default user and default groups.")
        try:
            login = sourcescfg["admin"]["login"]
            pwd = sourcescfg["admin"]["password"]
        except KeyError:
            if interactive:
                msg = "enter login and password of the initial manager account"
                login, pwd = manager_userpasswd(msg=msg, confirm=True)
            else:
                login, pwd = source["db-user"], source["db-password"]
        # sort for eid predicatability as expected in some server tests
        for group in sorted(BASE_GROUPS):
            cnx.create_entity("CWGroup", name=group)
        admin = create_user(cnx, login, pwd, "managers")
        cnx.execute(
            "SET X owned_by U WHERE X is IN (CWGroup,CWSource), U eid %(u)s",
            {"u": admin.eid},
        )
        cnx.commit()
    repo.shutdown()
    # re-login using the admin user
    config._cubes = None  # avoid assertion error
    repo = get_repository(config=config)
    # replace previous schema by the new repo's one. This is necessary so that we give the proper
    # schema to `initialize_schema` above since it will initialize .eid attribute of schema elements
    schema = repo.schema
    with connect(repo, login, password=pwd) as cnx:
        with cnx.security_enabled(False, False):
            repo.system_source.eid = ssource.eid  # redo this manually
            handler = config.migration_handler(
                schema, interactive=False, cnx=cnx, repo=repo
            )
            # serialize the schema
            initialize_schema(config, schema, handler)
            # yoo !
            cnx.commit()
            repo.system_source.init_creating()
            cnx.commit()
    repo.shutdown()
    # restore initial configuration
    config.creating = False
    config.consider_user_state = True
    # (drop instance attribute to get back to class attribute)
    del config.cubicweb_appobject_path
    del config.cube_appobject_path
    print(f"-> database for instance {config.appid} initialized.")


def initialize_schema(config, schema, mhandler, event="create"):
    from cubicweb.server.schemaserial import serialize_schema

    cnx = mhandler.cnx
    cubes = config.cubes()
    # deactivate every hooks but those responsible to set metadata
    # so, NO INTEGRITY CHECKS are done, to have quicker db creation.
    # Active integrity is kept else we may pb such as two default
    # workflows for one entity type.
    with cnx.deny_all_hooks_but("metadata", "activeintegrity"):
        # execute cubicweb's pre<event> script
        mhandler.cmd_exec_event_script(f"pre{event}")
        # execute cubes pre<event> script if any
        for cube in reversed(cubes):
            mhandler.cmd_exec_event_script(f"pre{event}", cube)
        # execute instance's pre<event> script (useful in tests)
        mhandler.cmd_exec_event_script(f"pre{event}", apphome=True)
        # enter instance'schema into the database
        serialize_schema(cnx, schema)
        cnx.commit()
        # execute cubicweb's post<event> script
        mhandler.cmd_exec_event_script(f"post{event}")
        # execute cubes'post<event> script if any
        for cube in reversed(cubes):
            mhandler.cmd_exec_event_script(f"post{event}", cube)
        # execute instance's post<event> script (useful in tests)
        mhandler.cmd_exec_event_script(f"post{event}", apphome=True)


# sqlite'stored procedures have to be registered at connection opening time

# add to this set relations which should have their add security checking done
# *BEFORE* adding the actual relation (done after by default)
BEFORE_ADD_RELATIONS = set(("owned_by",))

# add to this set relations which should have their add security checking done
# *at COMMIT TIME* (done after by default)
ON_COMMIT_ADD_RELATIONS = set(())

# available sources registry
SOURCE_TYPES = {
    "native": LazyObject("cubicweb.server.sources.native", "NativeSQLSource"),
    "datafeed": LazyObject("cubicweb.server.sources.datafeed", "DataFeedSource"),
    "ldapfeed": LazyObject("cubicweb.server.sources.ldapfeed", "LDAPFeedSource"),
}



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Integrity checking tool for instances:

* integrity of a CubicWeb repository. Hum actually only the system database is
  checked.
"""

import sys
from datetime import datetime

from logilab.common.shellutils import ProgressBar
from yams.constraints import UniqueConstraint

from cubicweb.schema import PURE_VIRTUAL_RTYPES, VIRTUAL_RTYPES, UNIQUE_CONSTRAINTS
from cubicweb.server.schema2sql import iter_unique_index_names, build_index_name
from cubicweb.server.sqlutils import SQL_PREFIX
from cubicweb.toolsutils import underline_title


def notify_fixed(fix):
    if fix:
        sys.stderr.write(" [FIXED]")
    sys.stderr.write("\n")


def has_eid(cnx, sqlcursor, eid, eids):
    """return true if the eid is a valid eid"""
    if eid in eids:
        return eids[eid]
    sqlcursor.execute(f"SELECT type FROM entities WHERE eid={eid}")
    try:
        etype = sqlcursor.fetchone()[0]
    except Exception:
        eids[eid] = False
        return False
    if etype not in cnx.vreg.schema:
        eids[eid] = False
        return False
    sqlcursor.execute(f"SELECT * FROM {SQL_PREFIX}{etype} WHERE {SQL_PREFIX}eid={eid}")
    result = sqlcursor.fetchall()
    if len(result) == 0:
        eids[eid] = False
        return False
    elif len(result) > 1:
        msg = (
            "  More than one entity with eid %s exists in source!\n"
            "  WARNING : Unable to fix this, do it yourself!\n"
        )
        sys.stderr.write(msg % eid)
    eids[eid] = True
    return True


# XXX move to yams?
def etype_fti_containers(eschema, _done=None):
    if _done is None:
        _done = set()
    _done.add(eschema)
    containers = tuple(eschema.fulltext_containers())
    if containers:
        for rschema, target in containers:
            if target == "object":
                targets = rschema.objects(eschema)
            else:
                targets = rschema.subjects(eschema)
            for targeteschema in targets:
                if targeteschema in _done:
                    continue
                _done.add(targeteschema)
                yield from etype_fti_containers(targeteschema, _done)
    else:
        yield eschema


def reindex_entities(schema, cnx, withpb=True, etypes=None):
    """reindex all entities in the repository"""
    # deactivate modification_date hook since we don't want them
    # to be updated due to the reindexation
    repo = cnx.repo
    dbhelper = repo.system_source.dbhelper
    cursor = cnx.cnxset.cu
    if not dbhelper.has_fti_table(cursor):
        print("no text index table")
        dbhelper.init_fti(cursor)
    repo.system_source.do_fti = True  # ensure full-text indexation is activated
    if etypes is None:
        print("Reindexing entities")
        etypes = set()
        for eschema in schema.entities():
            if eschema.final:
                continue
            indexable_attrs = tuple(eschema.indexable_attributes())  # generator
            if not indexable_attrs:
                continue
            for container in etype_fti_containers(eschema):
                etypes.add(container)
        # clear fti table first
        cnx.system_sql(f"DELETE FROM {dbhelper.fti_table}")
    else:
        print(
            f"Reindexing entities of type {', '.join(sorted(str(e) for e in etypes))}"
        )
        # clear fti table first. Use subquery for sql compatibility
        cnx.system_sql(
            "DELETE FROM %s WHERE EXISTS(SELECT 1 FROM ENTITIES "
            "WHERE eid=%s AND type IN (%s))"
            % (
                dbhelper.fti_table,
                dbhelper.fti_uid_attr,
                ",".join("'%s'" % etype for etype in etypes),
            )
        )
    if withpb:
        pb = ProgressBar(len(etypes) + 1)
        pb.update()
    # reindex entities by generating rql queries which set all indexable
    # attribute to their current value
    source = repo.system_source
    for eschema in etypes:
        etype_class = cnx.vreg["etypes"].etype_class(str(eschema))
        for rset in etype_class.cw_fti_index_rql_limit(cnx):
            source.fti_index_entities(cnx, rset.entities())
            # clear entity cache to avoid high memory consumption on big tables
            cnx.drop_entity_cache()
        if withpb:
            pb.update()
    if withpb:
        pb.finish()


_CHECKERS = {}


def _checker(func):
    """Decorator to register a function as a checker for check()."""
    fname = func.__name__
    prefix = "check_"
    assert fname.startswith(prefix), f"cannot register {func} as a checker"
    _CHECKERS[fname[len(prefix) :]] = func
    return func


@_checker
def check_schema(schema, cnx, eids, fix=1):
    """check serialized schema"""
    print("Checking serialized schema")
    rql = (
        "Any COUNT(X),RN,SN,ON,CTN GROUPBY RN,SN,ON,CTN ORDERBY 1 "
        "WHERE X is CWConstraint, R constrained_by X, "
        "R relation_type RT, RT name RN, R from_entity ST, ST name SN, "
        "R to_entity OT, OT name ON, X cstrtype CT, CT name CTN"
    )
    for count, rn, sn, on, cstrname in cnx.execute(rql):
        if count == 1:
            continue
        if cstrname in UNIQUE_CONSTRAINTS:
            print(
                "ERROR: got %s %r constraints on relation %s.%s.%s"
                % (count, cstrname, sn, rn, on)
            )
            if fix:
                print("dunno how to fix, do it yourself")


@_checker
def check_text_index(schema, cnx, eids, fix=1):
    """check all entities registered in the text index"""
    print("Checking text index")
    msg = (
        "  Entity with eid %s exists in the text index but not in any "
        "entity type table (autofix will remove from text index)"
    )
    cursor = cnx.system_sql("SELECT uid FROM appears;")
    for row in cursor.fetchall():
        eid = row[0]
        if not has_eid(cnx, cursor, eid, eids):
            sys.stderr.write(msg % eid)
            if fix:
                cnx.system_sql(f"DELETE FROM appears WHERE uid={eid};")
            notify_fixed(fix)


@_checker
def check_entities(schema, cnx, eids, fix=1):
    """check all entities registered in the repo system table"""
    print("Checking entities system table")
    # system table but no source
    msg = (
        '  Entity %s with eid %s exists in "entities" table but not in any '
        "entity type table (autofix will delete the entity)"
    )
    cursor = cnx.system_sql("SELECT eid,type FROM entities;")
    for row in cursor.fetchall():
        eid, etype = row
        if not has_eid(cnx, cursor, eid, eids):
            sys.stderr.write(msg % (etype, eid))
            if fix:
                cnx.system_sql(f"DELETE FROM entities WHERE eid={eid};")
            notify_fixed(fix)
    # source in entities, but no relation cw_source
    # XXX this (get_versions) requires a second connection to the db when we already have one open
    cursor = cnx.system_sql(
        "SELECT e.eid FROM entities as e, cw_CWSource as s "
        "WHERE NOT EXISTS(SELECT 1 FROM cw_source_relation as cs "
        "  WHERE cs.eid_from=e.eid) "
        "ORDER BY e.eid"
    )
    msg = "  Entity with eid %s is missing relation cw_source (autofix will create the relation)\n"
    for row in cursor.fetchall():
        sys.stderr.write(msg % row[0])
    if fix:
        cnx.system_sql(
            "INSERT INTO cw_source_relation (eid_from, eid_to) "
            "SELECT e.eid, s.cw_eid FROM entities as e, cw_CWSource as s "
            "WHERE s.cw_name='system' AND NOT EXISTS(SELECT 1 FROM cw_source_relation as cs "
            "  WHERE cs.eid_from=e.eid)"
        )
        notify_fixed(True)
    # inconsistencies for 'is'
    msg = '  %s #%s is missing relation "is" (autofix will create the relation)\n'
    cursor = cnx.system_sql(
        "SELECT e.type, e.eid FROM entities as e, cw_CWEType as s "
        "WHERE s.cw_name=e.type AND NOT EXISTS(SELECT 1 FROM is_relation as cs "
        "  WHERE cs.eid_from=e.eid AND cs.eid_to=s.cw_eid) "
        "ORDER BY e.eid"
    )
    for row in cursor.fetchall():
        sys.stderr.write(msg % tuple(row))
    if fix:
        cnx.system_sql(
            "INSERT INTO is_relation (eid_from, eid_to) "
            "SELECT e.eid, s.cw_eid FROM entities as e, cw_CWEType as s "
            "WHERE s.cw_name=e.type AND NOT EXISTS(SELECT 1 FROM is_relation as cs "
            "  WHERE cs.eid_from=e.eid AND cs.eid_to=s.cw_eid)"
        )
        notify_fixed(True)
    # inconsistencies for 'is_instance_of'
    msg = '  %s #%s is missing relation "is_instance_of" (autofix will create the relation)\n'
    cursor = cnx.system_sql(
        "SELECT e.type, e.eid FROM entities as e, cw_CWEType as s "
        "WHERE s.cw_name=e.type AND NOT EXISTS(SELECT 1 FROM is_instance_of_relation as cs "
        "  WHERE cs.eid_from=e.eid AND cs.eid_to=s.cw_eid) "
        "ORDER BY e.eid"
    )
    for row in cursor.fetchall():
        sys.stderr.write(msg % tuple(row))
    if fix:
        cnx.system_sql(
            "INSERT INTO is_instance_of_relation (eid_from, eid_to) "
            "SELECT e.eid, s.cw_eid FROM entities as e, cw_CWEType as s "
            "WHERE s.cw_name=e.type AND NOT EXISTS(SELECT 1 FROM is_instance_of_relation as cs "
            "  WHERE cs.eid_from=e.eid AND cs.eid_to=s.cw_eid)"
        )
        notify_fixed(True)
    print("Checking entities tables")
    msg = (
        '  Entity with eid %s exists in the %s table but not in "entities" '
        "table (autofix will delete the entity)"
    )
    for eschema in schema.entities():
        if eschema.final:
            continue
        table = SQL_PREFIX + eschema.type
        column = SQL_PREFIX + "eid"
        cursor = cnx.system_sql(f"SELECT {column} FROM {table};")
        for row in cursor.fetchall():
            eid = row[0]
            # eids is full since we have fetched everything from the entities table,
            # no need to call has_eid
            if eid not in eids or not eids[eid]:
                sys.stderr.write(msg % (eid, eschema.type))
                if fix:
                    cnx.system_sql(f"DELETE FROM {table} WHERE {column}={eid};")
                notify_fixed(fix)


def bad_related_msg(rtype, target, eid, fix):
    msg = (
        "  A relation %(rtype)s with %(target)s eid %(eid)d exists but "
        "entity #(eid)d does not exist"
    )
    sys.stderr.write(msg % {"rtype": rtype, "target": target, "eid": eid})
    notify_fixed(fix)


def bad_inlined_msg(rtype, parent_eid, eid, fix):
    msg = (
        "  An inlined relation %s from %s to %s exists but the latter "
        "entity does not exist"
    )
    sys.stderr.write(msg % (rtype, parent_eid, eid))
    notify_fixed(fix)


@_checker
def check_relations(schema, cnx, eids, fix=1):
    """check that eids referenced by relations are registered in the repo system
    table
    """
    print("Checking relations")
    for rschema in schema.relations():
        if rschema.final or rschema.rule or rschema.type in PURE_VIRTUAL_RTYPES:
            continue
        if rschema.inlined:
            for subjtype in rschema.subjects():
                table = SQL_PREFIX + str(subjtype)
                column = SQL_PREFIX + str(rschema)
                sql = f"SELECT cw_eid,{column} FROM {table} WHERE {column} IS NOT NULL;"
                cursor = cnx.system_sql(sql)
                for row in cursor.fetchall():
                    parent_eid, eid = row
                    if not has_eid(cnx, cursor, eid, eids):
                        bad_inlined_msg(rschema, parent_eid, eid, fix)
                        if fix:
                            sql = "UPDATE {} SET {}=NULL WHERE {}={};".format(
                                table,
                                column,
                                column,
                                eid,
                            )
                            cnx.system_sql(sql)
            continue
        try:
            cursor = cnx.system_sql(f"SELECT eid_from FROM {rschema}_relation;")
        except Exception as ex:
            # usually because table doesn't exist
            print("ERROR", ex)
            continue
        for row in cursor.fetchall():
            eid = row[0]
            if not has_eid(cnx, cursor, eid, eids):
                bad_related_msg(rschema, "subject", eid, fix)
                if fix:
                    sql = f"DELETE FROM {rschema}_relation WHERE eid_from={eid};"
                    cnx.system_sql(sql)
        cursor = cnx.system_sql(f"SELECT eid_to FROM {rschema}_relation;")
        for row in cursor.fetchall():
            eid = row[0]
            if not has_eid(cnx, cursor, eid, eids):
                bad_related_msg(rschema, "object", eid, fix)
                if fix:
                    sql = f"DELETE FROM {rschema}_relation WHERE eid_to={eid};"
                    cnx.system_sql(sql)


@_checker
def check_mandatory_relations(schema, cnx, eids, fix=1):
    """check entities missing some mandatory relation"""
    print("Checking mandatory relations")
    msg = "%s #%s is missing mandatory %s relation %s (autofix will delete the entity)"
    for rschema in schema.relations():
        if (
            rschema.final
            or rschema in PURE_VIRTUAL_RTYPES
            or rschema in ("is", "is_instance_of")
        ):
            continue
        smandatory = set()
        omandatory = set()
        for rdef in rschema.relation_definitions.values():
            if rdef.cardinality[0] in "1+":
                smandatory.add(rdef.subject)
            if rdef.cardinality[1] in "1+":
                omandatory.add(rdef.object)
        for role, etypes in (("subject", smandatory), ("object", omandatory)):
            for etype in etypes:
                if role == "subject":
                    rql = f"Any X WHERE NOT X {rschema} Y, X is {etype}"
                else:
                    rql = f"Any X WHERE NOT Y {rschema} X, X is {etype}"
                for entity in cnx.execute(rql).entities():
                    sys.stderr.write(msg % (entity.cw_etype, entity.eid, role, rschema))
                    if fix:
                        entity.cw_delete()  # XXX this is BRUTAL!
                    notify_fixed(fix)


@_checker
def check_mandatory_attributes(schema, cnx, eids, fix=1):
    """check for entities stored in the system source missing some mandatory
    attribute
    """
    print("Checking mandatory attributes")
    msg = "%s #%s is missing mandatory attribute %s (autofix will delete the entity)"
    for rschema in schema.relations():
        if not rschema.final or rschema in VIRTUAL_RTYPES:
            continue
        for rdef in rschema.relation_definitions.values():
            if rdef.cardinality[0] in "1+":
                rql = (
                    'Any X WHERE X %s NULL, X is %s, X cw_source S, S name "system"'
                    % (rschema, rdef.subject)
                )
                for entity in cnx.execute(rql).entities():
                    sys.stderr.write(msg % (entity.cw_etype, entity.eid, rschema))
                    if fix:
                        entity.cw_delete()
                    notify_fixed(fix)


@_checker
def check_metadata(schema, cnx, eids, fix=1):
    """check entities has required metadata

    FIXME: rewrite using RQL queries ?
    """
    print("Checking metadata")
    cursor = cnx.system_sql("SELECT DISTINCT type FROM entities;")
    eidcolumn = SQL_PREFIX + "eid"
    msg = "  %s with eid %s has no %s (autofix will set it to now)"
    for (etype,) in cursor.fetchall():
        if etype not in cnx.vreg.schema:
            sys.stderr.write(f"entities table references unknown type {etype}\n")
            if fix:
                cnx.system_sql(
                    "DELETE FROM entities WHERE type = %(type)s", {"type": etype}
                )
            continue
        table = SQL_PREFIX + etype
        for rel, default in (
            ("creation_date", datetime.utcnow()),
            ("modification_date", datetime.utcnow()),
        ):
            column = SQL_PREFIX + rel
            cursor = cnx.system_sql(
                f"SELECT {eidcolumn} FROM {table} WHERE {column} is NULL"
            )
            for (eid,) in cursor.fetchall():
                sys.stderr.write(msg % (etype, eid, rel))
                if fix:
                    cnx.system_sql(
                        f"UPDATE {table} SET {column}=%(v)s WHERE {eidcolumn}={eid} ;",
                        {"v": default},
                    )
                notify_fixed(fix)


def check(repo, cnx, checks, reindex, fix, withpb=True):
    """check integrity of instance's repository,
    using given user and password to locally connect to the repository
    (no running cubicweb server needed)
    """
    # yo, launch checks
    if checks:
        eids_cache = {}
        with cnx.security_enabled(read=False, write=False):  # ensure no read security
            for check in checks:
                check_func = _CHECKERS[check]
                check_func(repo.schema, cnx, eids_cache, fix=fix)
        if fix:
            cnx.commit()
        else:
            print()
        if not fix:
            print("WARNING: Diagnostic run, nothing has been corrected")
    if reindex:
        cnx.rollback()
        reindex_entities(repo.schema, cnx, withpb=withpb)
        cnx.commit()


SYSTEM_INDEXES = {
    # see cw/server/sources/native.py
    "transactions_tx_time_idx": ("transactions", "tx_time"),
    "transactions_tx_user_idx": ("transactions", "tx_user"),
    "tx_entity_actions_txa_action_idx": ("tx_entity_actions", "txa_action"),
    "tx_entity_actions_txa_public_idx": ("tx_entity_actions", "txa_public"),
    "tx_entity_actions_eid_idx": ("tx_entity_actions", "txa_eid"),
    "tx_entity_actions_etype_idx": ("tx_entity_actions", "txa_etype"),
    "tx_entity_actions_tx_uuid_idx": ("tx_entity_actions", "tx_uuid"),
    "tx_relation_actions_txa_action_idx": ("tx_relation_actions", "txa_action"),
    "tx_relation_actions_txa_public_idx": ("tx_relation_actions", "txa_public"),
    "tx_relation_actions_eid_from_idx": ("tx_relation_actions", "eid_from"),
    "tx_relation_actions_eid_to_idx": ("tx_relation_actions", "eid_to"),
    "tx_relation_actions_tx_uuid_idx": ("tx_relation_actions", "tx_uuid"),
}


def expected_indexes(cnx):
    """Return a dictionary describing indexes expected by the schema {index name: (table, column)}.

    This doesn't include primary key indexes.
    """
    source = cnx.repo.system_source
    dbh = source.dbhelper
    schema = cnx.repo.schema
    schema_indexes = SYSTEM_INDEXES.copy()
    if source.dbdriver == "postgres":
        schema_indexes.update({"appears_words_idx": ("appears", "words")})
    else:
        schema_indexes.update(
            {
                "appears_uid": ("appears", "uid"),
                "appears_word_id": ("appears", "word_id"),
            }
        )
    for rschema in schema.relations():
        if rschema.rule or rschema in PURE_VIRTUAL_RTYPES:
            continue  # computed relation
        if rschema.final or rschema.inlined:
            for rdef in rschema.relation_definitions.values():
                table = f"cw_{rdef.subject}"
                column = f"cw_{rdef.relation_type}"
                if any(isinstance(cstr, UniqueConstraint) for cstr in rdef.constraints):
                    schema_indexes[dbh._index_name(table, column, unique=True)] = (
                        table,
                        [column],
                    )
                if rschema.inlined or rdef.indexed:
                    schema_indexes[dbh._index_name(table, column)] = (table, [column])
        else:
            table = f"{rschema}_relation"
            if source.dbdriver == "postgres":
                # index built after the primary key constraint
                schema_indexes[
                    build_index_name(table, ["eid_from", "eid_to"], "key_")
                ] = (table, ["eid_from", "eid_to"])
            schema_indexes[build_index_name(table, ["eid_from"], "idx_")] = (
                table,
                ["eid_from"],
            )
            schema_indexes[build_index_name(table, ["eid_to"], "idx_")] = (
                table,
                ["eid_to"],
            )
    for eschema in schema.entities():
        if eschema.final:
            continue
        table = f"cw_{eschema}"
        for columns, index_name in iter_unique_index_names(eschema):
            schema_indexes[index_name] = (table, columns)

    return schema_indexes


def database_indexes(cnx):
    """Return a set of indexes found in the database, excluding primary key indexes."""
    source = cnx.repo.system_source
    dbh = source.dbhelper
    if source.dbdriver == "postgres":

        def index_filter(idx):
            return not (
                idx.startswith("pg_")
                or "_pkey" in idx
                or "_p_key" in idx
                or idx.endswith("_key")
            )

    else:

        def index_filter(idx):
            return not idx.startswith("sqlite_")

    return {idx for idx in dbh.list_indices(cnx.cnxset.cu) if index_filter(idx)}


def check_indexes(cnx):
    """Check indexes of a system database: output missing expected indexes as well as unexpected ones.

    Return 0 if there is no differences, else 1.
    """
    schema_indexes = expected_indexes(cnx)
    db_indexes = database_indexes(cnx)

    missing_indexes = set(schema_indexes) - db_indexes
    if missing_indexes:
        print(underline_title("Missing indexes"))
        print("index expected by the schema but not found in the database:\n")
        missing = [
            "{0} ON {1[0]} {1[1]}".format(idx, schema_indexes[idx])
            for idx in missing_indexes
        ]
        print("\n".join(sorted(missing)))
        print()
        status = 1

    additional_indexes = db_indexes - set(schema_indexes)
    if additional_indexes:
        print(underline_title("Additional indexes"))
        print("index in the database but not expected by the schema:\n")
        print("\n".join(sorted(additional_indexes)))
        print()
        status = 1

    if not (missing_indexes or additional_indexes):
        print("Everything is Ok")
        status = 0

    return status



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""
.. autoclass:: cubicweb.dataimport.massive_store.MassiveObjectStore
"""
import logging
from collections import defaultdict
from itertools import chain
from uuid import uuid4

from cubicweb.dataimport import stores, pgstore
from cubicweb.server.schema2sql import eschema_sql_def


class MassiveObjectStore(stores.RQLObjectStore):
    """Store for massive import of data, with delayed insertion of meta data.

    WARNINGS:

    - This store may only be used with PostgreSQL for now, as it relies
      on the COPY FROM method, and on specific PostgreSQL tables to get all
      the indexes.

    - This store can only insert relations that are not inlined (i.e.,
      which do *not* have inlined=True in their definition in the schema),
      unless they are specified as entity attributes.

    It should be used as follows:

       store = MassiveObjectStore(cnx)
       eid_p = store.prepare_insert_entity('Person',
                                           cwuri=u'http://dbpedia.org/toto',
                                           name=u'Toto')
       eid_loc = store.prepare_insert_entity('Location',
                                             cwuri=u'http://geonames.org/11111',
                                             name=u'Somewhere')
       store.prepare_insert_relation(eid_p, 'lives_in', eid_loc)
       store.flush()
       ...
       store.commit()
       store.finish()

    Full-text indexation is not handled, you'll have to reindex the proper entity types by yourself
    if desired.
    """

    def __init__(
        self, cnx, slave_mode=False, eids_seq_range=10000, metagen=None, drop=True
    ):
        """Create a MassiveObject store, with the following arguments:

        - `cnx`, a connection to the repository
        - `metagen`, optional :class:`MetadataGenerator` instance
        - `eids_seq_range`: size of eid range reserved by the store for each batch
        """
        super(MassiveObjectStore, self).__init__(cnx)

        self.uuid = str(uuid4()).replace("-", "")
        self.slave_mode = slave_mode
        if metagen is None:
            metagen = stores.MetadataGenerator(cnx)
        self.metagen = metagen

        self.logger = logging.getLogger("dataimport.massive_store")
        self.sql = cnx.system_sql
        self.schema = cnx.vreg.schema
        self.default_values = get_default_values(self.schema)
        self.get_next_eid = lambda g=self._get_eid_gen(eids_seq_range): next(g)
        self._source_dbhelper = cnx.repo.system_source.dbhelper
        if drop:
            self._dbh = PGHelper(cnx)
        else:
            self._dbh = NoDropPGHelper(cnx)

        self._data_entities = defaultdict(list)
        self._data_relations = defaultdict(list)
        self._data_entities_to_update = defaultdict(list)
        self._initialized = {}

    def _get_eid_gen(self, eids_seq_range):
        """Function getting the next eid. This is done by preselecting
        a given number of eids from the 'entities_id_seq', and then
        storing them"""
        while True:
            last_eid = self._cnx.repo.system_source.create_eid(
                self._cnx, eids_seq_range
            )
            for eid in range(last_eid - eids_seq_range + 1, last_eid + 1):
                yield eid

    # master/slaves specific API

    def master_init(self, commit=True):
        """Initialize database for massive insertion.

        This is expected to be called once, by the master store in master/slaves configuration.
        """
        assert not self.slave_mode
        if self not in self._initialized:
            self.sql("DROP TABLE IF EXISTS cwmassive_initialized")
            self.sql(
                "CREATE TABLE cwmassive_initialized"
                "(retype text, type varchar(128), uuid varchar(32))"
            )
            self._initialized[self] = None
            if commit:
                self.commit()

    # SQL utilities #########################################################

    def _drop_metadata_constraints(self):
        """Drop constraints and indexes for the metadata tables.

        They will be recreated by the `finish` method.
        """
        rtypes = [
            rtype
            for rtype in self.metagen.meta_relations
            if not self.schema.relation_schema_for(rtype).final
        ]
        rtypes += ("is_instance_of", "is", "cw_source")
        for rtype in rtypes:
            self._dbh.drop_constraints(rtype + "_relation")
            self._dbh.drop_indexes(rtype + "_relation")
        # don't drop constraints for the entities table, the only one is the primary key's index on
        # eid and we want to keep it
        self._dbh.drop_indexes("entities")

    def restart_eid_sequence(self, start_eid):
        self.sql(
            self._cnx.repo.system_source.dbhelper.sql_restart_numrange(
                "entities_id_seq", initial_value=start_eid
            )
        )
        self._cnx.commit()

    # store api ################################################################

    def prepare_insert_entity(self, etype, **data):
        """Given an entity type, attributes and inlined relations, returns the inserted entity's
        eid.
        """
        if etype not in self._initialized:
            if not self.slave_mode:
                self.master_init(commit=False)
            tablename = f"cw_{etype.lower()}"
            tmp_tablename = f"{tablename}_{self.uuid}"
            self.sql(
                "INSERT INTO cwmassive_initialized VALUES (%(e)s, 'etype', %(uuid)s)",
                {"e": etype, "uuid": self.uuid},
            )
            attr_defs = eschema_sql_def(self._source_dbhelper, self.schema[etype])
            self.sql(
                "CREATE TABLE %s(%s);"
                % (
                    tmp_tablename,
                    ", ".join(
                        "cw_%s %s" % (column, sqltype) for column, sqltype in attr_defs
                    ),
                )
            )
            self._initialized[etype] = [attr for attr, _ in attr_defs]

        if "eid" not in data:
            # If eid is not given and the eids sequence is set, use the value from the sequence
            eid = self.get_next_eid()
            data["eid"] = eid
        self._data_entities[etype].append(data)
        return data["eid"]

    def prepare_insert_relation(self, eid_from, rtype, eid_to, **kwargs):
        """Insert into the database a  relation ``rtype`` between entities with eids ``eid_from``
        and ``eid_to``.

        Relation must not be inlined.
        """
        if rtype not in self._initialized:
            if not self.slave_mode:
                self.master_init(commit=False)
            assert not self._cnx.vreg.schema.relation_schema_for(rtype).inlined
            self._initialized[rtype] = None
            tablename = f"{rtype.lower()}_relation"
            tmp_tablename = f"{tablename}_{self.uuid}"
            self.sql(
                "INSERT INTO cwmassive_initialized VALUES (%(r)s, 'rtype', %(uuid)s)",
                {"r": rtype, "uuid": self.uuid},
            )
            self.sql(f"CREATE TABLE {tmp_tablename}(eid_from integer, eid_to integer)")
        self._data_relations[rtype].append({"eid_from": eid_from, "eid_to": eid_to})

    def prepare_update_entity(self, etype, eid, **data):
        """Given an entity type, eid, attributes and inlined relations, returns
        the updated entity's eid.
        """
        entity = self._cnx.entity_from_eid(eid)
        assert entity.cw_etype == etype, f"Trying to update with wrong type {etype}"
        initialized_key = f"update_{etype}"
        if initialized_key not in self._initialized:
            if not self.slave_mode:
                self.master_init(commit=False)
            tablename = f"cw_{etype.lower()}_to_update"
            tmp_tablename = f"{tablename}_{self.uuid}"
            self.sql(
                "INSERT INTO cwmassive_initialized VALUES (%(e)s, 'update_etype', %(uuid)s)",
                {"e": etype, "uuid": self.uuid},
            )
            attr_defs = eschema_sql_def(self._source_dbhelper, self.schema[etype])
            self.sql(
                "CREATE TABLE %s(%s);"
                % (
                    tmp_tablename,
                    ", ".join(
                        "cw_%s %s" % (column, sqltype) for column, sqltype in attr_defs
                    ),
                )
            )
            self._initialized[initialized_key] = [attr for attr, _ in attr_defs]

        # we do that to avoid a lot of warning :)
        completed_data = {attr: None for attr in self._initialized[initialized_key]}
        self._data_entities_to_update[etype].append(
            {**completed_data, **data, "eid": eid}
        )
        return eid

    def flush(self):
        """Flush the data"""
        self.flush_entities()
        self.flush_entities_to_update()
        self.flush_relations()

    def finish(self):
        """Remove temporary tables and columns."""
        try:
            self._finish()
            self._cnx.commit()
        except Exception:
            self._cnx.rollback()
            raise
        finally:
            # delete the meta data table
            self.sql("DROP TABLE IF EXISTS cwmassive_initialized")
            self.commit()
            self._initialized = {}

    def _finish(self):
        """Remove temporary tables and columns."""
        assert (
            not self.slave_mode
        ), "finish method should only be called by the master store"
        self.logger.info("Start cleaning")
        # Get all the initialized etypes/rtypes
        if self._dbh.table_exists("cwmassive_initialized"):
            cu = self.sql("SELECT retype, type, uuid FROM cwmassive_initialized")
            entities = defaultdict(list)
            relations = defaultdict(list)
            entities_to_update = defaultdict(list)
            for retype, _type, uuid in cu.fetchall():
                if _type == "rtype":
                    relations[retype].append(uuid)
                elif _type == "update_etype":
                    entities_to_update[retype].append(uuid)
                else:  # _type = 'etype'
                    entities[retype].append(uuid)
            # if there is some entities to insert, delete constraint on metadata tables once for all
            if entities:
                self._drop_metadata_constraints()
            # get back entity to update data from the temporary tables
            for etype, uuids in entities_to_update.items():
                tablename = f"cw_{etype.lower()}"
                attr_defs = eschema_sql_def(self._source_dbhelper, self.schema[etype])
                set_parts = []
                for attr, _ in attr_defs:
                    if attr == "eid":
                        continue
                    set_parts.append(
                        f"cw_{attr} = COALESCE(TMP.cw_{attr}, {tablename}.cw_{attr})"
                    )

                self._dbh.drop_constraints(tablename)
                for uuid in uuids:
                    tmp_tablename = f"{tablename}_to_update_{uuid}"
                    self.sql(
                        f"""
                        UPDATE {tablename}
                        SET {",".join(set_parts)}
                        FROM {tmp_tablename} as TMP
                        WHERE {tablename}.cw_eid = TMP.cw_eid
                        """
                    )
                    self._tmp_data_cleanup(tmp_tablename, f"update_{etype}", uuid)
            # get back entity data from the temporary tables
            for etype, uuids in entities.items():
                tablename = f"cw_{etype.lower()}"
                attr_defs = eschema_sql_def(self._source_dbhelper, self.schema[etype])
                columns = ",".join(f"cw_{attr}" for attr, _ in attr_defs)
                self._dbh.drop_constraints(tablename)
                self._dbh.drop_indexes(tablename)
                for uuid in uuids:
                    tmp_tablename = f"{tablename}_{uuid}"
                    self.sql(
                        "INSERT INTO %(table)s(%(columns)s) "
                        "SELECT %(columns)s FROM %(tmp_table)s"
                        % {
                            "table": tablename,
                            "tmp_table": tmp_tablename,
                            "columns": columns,
                        }
                    )
                    self._insert_etype_metadata(etype, tmp_tablename)
                    self._tmp_data_cleanup(tmp_tablename, etype, uuid)
            # get back relation data from the temporary tables
            for rtype, uuids in relations.items():
                tablename = f"{rtype.lower()}_relation"
                self._dbh.drop_constraints(tablename)
                self._dbh.drop_indexes(tablename)
                for uuid in uuids:
                    tmp_tablename = f"{tablename}_{uuid}"
                    self.fill_relation_table(tablename, tmp_tablename)
                    self._tmp_data_cleanup(tmp_tablename, rtype, uuid)
        # restore all deleted indexes and constraints
        self._dbh.restore_indexes_and_constraints()

    def _insert_etype_metadata(self, etype, tmp_tablename):
        """Massive insertion of meta data for `etype`, with new entities in `tmp_tablename`."""
        # insert standard metadata relations
        for rtype, eid in self.metagen.base_etype_rels(etype).items():
            self.fill_meta_relation_table(tmp_tablename, rtype, eid)
        # insert cw_source, is and is_instance_of relations (normally handled by the system source)
        self.fill_meta_relation_table(
            tmp_tablename, "cw_source", self.metagen.source.eid
        )
        eschema = self.schema[etype]
        self.fill_meta_relation_table(tmp_tablename, "is", eschema.eid)
        for parent_eschema in chain(eschema.ancestors(), [eschema]):
            self.fill_meta_relation_table(
                tmp_tablename, "is_instance_of", parent_eschema.eid
            )
        self.fill_entities_table(etype, tmp_tablename)

    def fill_entities_table(self, etype, tmp_tablename):
        # finally insert records into the entities table
        self.sql(
            "INSERT INTO entities(eid, type) "
            "SELECT cw_eid, '%s' FROM %s "
            "WHERE NOT EXISTS (SELECT 1 FROM entities WHERE eid=cw_eid)"
            % (etype, tmp_tablename)
        )

    def fill_relation_table(self, tablename, tmp_tablename):
        # XXX no index on the original relation table, EXISTS subquery may be sloooow
        self.sql(
            "INSERT INTO %(table)s(eid_from, eid_to) SELECT DISTINCT "
            "T.eid_from, T.eid_to FROM %(tmp_table)s AS T "
            "WHERE NOT EXISTS (SELECT 1 FROM %(table)s AS TT WHERE "
            "TT.eid_from=T.eid_from AND TT.eid_to=T.eid_to);"
            % {"table": tablename, "tmp_table": tmp_tablename}
        )

    def fill_meta_relation_table(self, tmp_tablename, rtype, eid_to):
        self.sql(
            "INSERT INTO %s_relation(eid_from, eid_to) SELECT cw_eid, %s FROM %s "
            "WHERE NOT EXISTS (SELECT 1 FROM entities WHERE eid=cw_eid)"
            % (rtype, eid_to, tmp_tablename)
        )

    def _tmp_data_cleanup(self, tmp_tablename, ertype, uuid):
        """Drop temporary relation table and record from cwmassive_initialized."""
        self.sql(f"DROP TABLE {tmp_tablename}")
        self.sql(
            "DELETE FROM cwmassive_initialized "
            "WHERE retype = %(rtype)s AND uuid = %(uuid)s",
            {"rtype": ertype, "uuid": uuid},
        )

    # FLUSH #################################################################

    def flush_relations(self):
        """Flush the relations data from in-memory structures to a temporary table."""
        for rtype, data in self._data_relations.items():
            if not data:
                # There is no data for these etype for this flush round.
                continue
            buf = pgstore._create_copyfrom_buffer(data, ("eid_from", "eid_to"))
            cursor = self._cnx.cnxset.cu
            tablename = f"{rtype.lower()}_relation"
            tmp_tablename = f"{tablename}_{self.uuid}"
            cursor.copy_from(
                buf, tmp_tablename, null="NULL", columns=("eid_from", "eid_to")
            )
            # Clear data cache
            self._data_relations[rtype] = []

    def flush_entities(self):
        """Flush the entities data from in-memory structures to a temporary table."""
        metagen = self.metagen
        for etype, data in self._data_entities.items():
            if not data:
                # There is no data for these etype for this flush round.
                continue
            attrs = self._initialized[etype]
            _base_data = dict.fromkeys(attrs)
            _base_data.update(self.default_values[etype])
            _base_data.update(metagen.base_etype_attrs(etype))
            _data = []
            for d in data:
                # do this first on `d`, because it won't fill keys associated to None as provided by
                # `_base_data`
                metagen.init_entity_attrs(etype, d["eid"], d)
                # XXX warn/raise if there is some key not in attrs?
                _d = _base_data.copy()
                _d.update(d)
                _data.append(_d)
            buf = pgstore._create_copyfrom_buffer(_data, attrs)
            tablename = f"cw_{etype.lower()}"
            tmp_tablename = f"{tablename}_{self.uuid}"
            columns = [f"cw_{attr}" for attr in attrs]
            cursor = self._cnx.cnxset.cu
            cursor.copy_from(buf, tmp_tablename, null="NULL", columns=columns)
            # Clear data cache
            self._data_entities[etype] = []

    def flush_entities_to_update(self):
        """Flush the entities to update data from in-memory structures to a temporary table."""
        for etype, data in self._data_entities_to_update.items():
            if not data:
                # There is no data for these etype for this flush round.
                continue
            attrs = self._initialized[f"update_{etype}"]
            buf = pgstore._create_copyfrom_buffer(data, attrs)
            tablename = f"cw_{etype.lower()}_to_update"
            tmp_tablename = f"{tablename}_{self.uuid}"
            columns = [f"cw_{attr}" for attr in attrs]
            cursor = self._cnx.cnxset.cu
            cursor.copy_from(buf, tmp_tablename, null="NULL", columns=columns)
            # Clear data cache
            self._data_entities_to_update[etype] = []


def get_default_values(schema):
    """analyzes yams ``schema`` and returns the list of default values.

    The returned value is a dictionary mapping entity types to a
    sub-dictionnaries mapping attribute names -> default values.
    """
    default_values = {}
    # iterates on all entity types
    for eschema in schema.entities():
        # for each entity type, iterates on attribute definitions
        default_values[eschema.type] = eschema_constraints = {}
        for rschema, _ in eschema.attribute_definitions():
            # for each attribute, if a size constraint is found,
            # append it to the size constraint list
            if eschema.default(rschema.type) is not None:
                eschema_constraints[rschema.type] = eschema.default(rschema.type)
    return default_values


class PGHelper:
    """This class provides some helper methods to manipulate a postgres database metadata (index and
    constraints).
    """

    def __init__(self, cnx):
        self.sql = cnx.system_sql
        # Deals with pg schema, see #3216686
        pg_schema = cnx.repo.config.system_source_config.get("db-namespace") or "public"
        self.pg_schema = pg_schema

    def drop_indexes(self, tablename):
        """Drop indexes and constraints, storing them in a table for later restore."""
        # Create a table to save the constraints, it allows reloading even after crash
        self.sql(
            "CREATE TABLE IF NOT EXISTS cwmassive_constraints(sql TEXT, insert_order SERIAL)"
        )
        indexes = self.table_indexes(tablename)
        for name, query in indexes.items():
            self.sql(
                "INSERT INTO cwmassive_constraints(sql) VALUES (%(sql)s)",
                {"sql": query},
            )
            self.sql(f"DROP INDEX {name}")

    def drop_constraints(self, tablename):
        self.sql(
            "CREATE TABLE IF NOT EXISTS cwmassive_constraints(sql TEXT, insert_order SERIAL)"
        )
        constraints = self.table_constraints(tablename)
        for name, query in constraints.items():
            self.sql(
                "INSERT INTO cwmassive_constraints(sql) VALUES (%(sql)s)",
                {"sql": query},
            )
            self.sql(f"ALTER TABLE {tablename} DROP CONSTRAINT {name}")

    def restore_indexes_and_constraints(self):
        """Restore indexes and constraints."""
        if not self.table_exists("cwmassive_constraints"):
            return
        cu = self.sql(
            "SELECT sql, insert_order FROM cwmassive_constraints "
            "ORDER BY insert_order DESC"
        )
        for query, order in cu.fetchall():
            self.sql(query)
            self.sql(
                "DELETE FROM cwmassive_constraints WHERE insert_order=%(order)s",
                {"order": order},
            )
        self.sql("DROP TABLE cwmassive_constraints")

    def table_exists(self, tablename):
        """Return True if the given table already exists in the database."""
        cu = self.sql(
            "SELECT 1 from information_schema.tables "
            "WHERE table_name=%(t)s AND table_schema=%(s)s",
            {"t": tablename, "s": self.pg_schema},
        )
        return bool(cu.fetchone())

    def table_indexes(self, tablename):
        """Return a dictionary of indexes {index name: index sql}, constraints included."""
        indexes = {}
        for name in self._index_names(tablename):
            indexes[name] = self._index_sql(name)
        return indexes

    def table_constraints(self, tablename):
        """Return a dictionary of constraints {constraint name: constraint sql}."""
        constraints = {}
        for name in self._constraint_names(tablename):
            query = self._constraint_sql(name)
            constraints[name] = f"ALTER TABLE {tablename} ADD CONSTRAINT {name} {query}"
        return constraints

    def _index_names(self, tablename):
        """Return the names of all indexes in the given table (including constraints.)"""
        cu = self.sql(
            "SELECT c.relname FROM pg_catalog.pg_class c "
            "JOIN pg_catalog.pg_index i ON i.indexrelid = c.oid "
            "JOIN pg_catalog.pg_class c2 ON i.indrelid = c2.oid "
            "LEFT JOIN pg_catalog.pg_user u ON u.usesysid = c.relowner "
            "LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace "
            "WHERE c.relkind IN ('i','') "
            " AND c2.relname = %(t)s "
            " AND i.indisprimary = FALSE "
            " AND n.nspname NOT IN ('pg_catalog', 'pg_toast') "
            " AND pg_catalog.pg_table_is_visible(c.oid);",
            {"t": tablename},
        )
        return [name for name, in cu.fetchall()]

    def _constraint_names(self, tablename):
        """Return the names of all constraints in the given table."""
        cu = self.sql(
            "SELECT i.conname FROM pg_catalog.pg_class c "
            "JOIN pg_catalog.pg_constraint i ON i.conrelid = c.oid "
            "JOIN pg_catalog.pg_class c2 ON i.conrelid=c2.oid "
            "LEFT JOIN pg_catalog.pg_user u ON u.usesysid = c.relowner "
            "LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace "
            "WHERE c2.relname = %(t)s "
            "AND n.nspname NOT IN ('pg_catalog', 'pg_toast') "
            "AND pg_catalog.pg_table_is_visible(c.oid)",
            {"t": tablename},
        )
        return [name for name, in cu.fetchall()]

    def _index_sql(self, name):
        """Return the SQL to be used to recreate the index of the given name."""
        return self.sql(
            "SELECT pg_get_indexdef(c.oid) FROM pg_catalog.pg_class c "
            "LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace "
            "WHERE c.relname = %(r)s AND n.nspname=%(n)s",
            {"r": name, "n": self.pg_schema},
        ).fetchone()[0]

    def _constraint_sql(self, name):
        """Return the SQL to be used to recreate the constraint."""
        return self.sql(
            "SELECT pg_get_constraintdef(c.oid) FROM pg_catalog.pg_constraint c "
            "LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.connamespace "
            "WHERE c.conname = %(r)s AND n.nspname=%(n)s",
            {"r": name, "n": self.pg_schema},
        ).fetchone()[0]


class NoDropPGHelper(PGHelper):
    """Custom PGHelper that doesn't drop any index nor constraint.
    Instead of dropping the constraints, they are disabled, and restored at
    the end of the import.

    It's usually faster when a table already contains data that you want to
    keep.
    """

    def drop_indexes(self, tablename):
        pass

    def drop_constraints(self, tablename):
        """Disable constraints on the given table."""
        self.sql(
            "CREATE TABLE IF NOT EXISTS cwmassive_constraints(sql TEXT, insert_order SERIAL)"
        )

        self.sql(f"ALTER TABLE {tablename} DISABLE TRIGGER ALL;")

        self.sql(
            "INSERT INTO cwmassive_constraints(sql) VALUES (%(sql)s)",
            {"sql": f"ALTER TABLE {tablename} ENABLE TRIGGER ALL;"},
        )



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Simple cryptographic routines, based on python-crypto."""


import pickle
from base64 import b64encode, b64decode

from Cryptodome.Cipher import Blowfish

_CYPHERERS = {}


def _cypherer(seed):
    if isinstance(seed, str):
        seed = seed.encode("utf-8")
    try:
        return _CYPHERERS[seed]
    except KeyError:
        _CYPHERERS[seed] = Blowfish.new(seed, Blowfish.MODE_ECB)
        return _CYPHERERS[seed]


def encrypt(data, seed):
    string = pickle.dumps(data)
    string = string + b"*" * (8 - len(string) % 8)
    string = b64encode(_cypherer(seed).encrypt(string))
    return string.decode("utf-8")


def decrypt(string, seed):
    string = string.encode("utf-8")
    # pickle ignores trailing characters so we do not need to strip them off
    string = _cypherer(seed).decrypt(b64decode(string))
    return pickle.loads(string)



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

import collections
from logging import getLogger
from contextlib import contextmanager

LOGGER = getLogger("cubicweb")


def _compute_impacted_entities_and_relations_to_delete(
    cnx, eid, etype, to_delete, rels_to_delete, to_update
):
    if eid in to_delete.get(etype, []):
        # avoid loops
        return
    to_delete.setdefault(etype, set()).add(eid)

    eschema = cnx.repo.schema[etype]

    for rschema, related_etypes, role in eschema.relation_definitions():
        if rschema.type == "identity":
            continue
        composites = []
        non_composites = []
        for related_etype in related_etypes:
            rdef = rschema.role_relation_definition(eschema, related_etype, role)
            if role == rdef.composite:
                composites.append(related_etype.type)
            else:
                non_composites.append(related_etype.type)

        if composites and non_composites:
            raise NotImplementedError

        if non_composites and role == "subject" and rschema.inlined:
            continue

        if non_composites and role == "object" and rschema.inlined:
            for related_etype in non_composites:
                to_update.setdefault((related_etype, rschema.type), set()).add(eid)
        elif not rschema.inlined:
            rels_to_delete.setdefault((rschema.type, role), set()).add(eid)

        if composites:
            rql = "Any Y, E WHERE Y is ET, ET name E, X eid %(eid)s"
            if role == "subject":
                rql = ",".join([rql, f"X {rschema.type} Y"])
            else:
                rql = ",".join([rql, f"Y {rschema.type} X"])
            for related_eid, related_etype in cnx.execute(
                rql, {"eid": eid}, build_descr=False
            ):
                _compute_impacted_entities_and_relations_to_delete(
                    cnx,
                    related_eid,
                    related_etype,
                    to_delete,
                    rels_to_delete,
                    to_update,
                )


def _run_fast_drop_entities(cnx, to_update, rels_to_delete, to_delete):
    for (etype, rtype), eids in to_update.items():
        cnx.system_sql(
            "UPDATE cw_{0} SET cw_{1} = NULL WHERE cw_{1} IN ({2})".format(
                etype, rtype, ",".join(str(e) for e in eids)
            )
        )

    for (rtype, role), eids in rels_to_delete.items():
        cnx.system_sql(
            "DELETE FROM {}_relation WHERE {} IN ({})".format(
                rtype,
                "eid_from" if role == "subject" else "eid_to",
                ",".join(str(e) for e in eids),
            )
        )

    deleted = set()
    for etype, eids in to_delete.items():
        deleted |= eids
        cnx.system_sql(
            f"DELETE FROM cw_{etype} WHERE cw_eid IN ({','.join(str(e) for e in eids)})"
        )
    if deleted:
        eids = ",".join(str(e) for e in deleted)
        cnx.system_sql(f"DELETE FROM appears WHERE uid IN ({eids})")
        cnx.system_sql(f"DELETE FROM entities WHERE eid IN ({eids})")


@contextmanager
def disable_constraints_check(cnx):
    # disable all triggers and constraint check
    if cnx.repo.system_source.dbdriver == "postgres":
        cnx.system_sql("SET session_replication_role = replica;")
    elif cnx.repo.system_source.dbdriver == "sqlite":
        cnx.system_sql("PRAGMA foreign_keys = ON;")
    else:
        LOGGER.warning(
            "Unknown db-driver : %s."
            "The triggers are **NOT** disable during deletion and "
            "the function might not be as fast as expected",
            cnx.repo.system_source.dbdriver,
        )

    try:
        yield cnx
    except Exception:
        cnx.rollback()
        raise
    finally:  # restore all triggers and constraint check
        if cnx.repo.system_source.dbdriver == "postgres":
            cnx.system_sql("SET session_replication_role = DEFAULT;")
        elif cnx.repo.system_source.dbdriver == "sqlite":
            cnx.system_sql("PRAGMA foreign_keys = OFF;")


def fast_drop_entities(
    rset, autocommit=True, disable_triggers_and_constraint_check=True
):
    """Delete entities faster using raw SQL and without executing hooks and
    with or without constraint check
    """
    cnx = rset.req
    to_delete = collections.OrderedDict()
    rels_to_delete = collections.OrderedDict()
    to_update = collections.OrderedDict()
    for entity in rset.entities():
        _compute_impacted_entities_and_relations_to_delete(
            cnx, entity.eid, entity.cw_etype, to_delete, rels_to_delete, to_update
        )

    if disable_triggers_and_constraint_check:
        with disable_constraints_check(cnx) as cnx:
            _run_fast_drop_entities(cnx, to_update, rels_to_delete, to_delete)
    else:
        _run_fast_drop_entities(cnx, to_update, rels_to_delete, to_delete)

    if autocommit:
        cnx.commit()



if repo.system_source.dbdriver == "postgres":
    # the postgresql GROUP_CONCAT function has been changed.
    # we need to reset all the function so that change is taken into account
    install_custom_sql_scripts()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Usage: %s [OPTIONS] <instance id> <queries file>

Stress test a CubicWeb repository

OPTIONS:
  -h / --help
     Display this help message and exit.

  -u / --user <user>
     Connect as <user> instead of being prompted to give it.
  -p / --password <password>
     Automatically give <password> for authentication instead of being prompted
     to give it.

  -n / --nb-times <num>
     Repeat queries <num> times.
  -t / --nb-threads <num>
     Execute queries in <num> parallel threads.
  -P / --profile <prof_file>
     dumps profile results (hotshot) in <prof_file>
  -o / --report-output <filename>
     Write profiler report into <filename> rather than on stdout

Copyright (c) 2003-2023 LOGILAB S.A. (Paris, FRANCE), license is LGPL v2.
https://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
import getopt
import os
import sys
import threading
import traceback
from getpass import getpass
from os.path import basename
from time import process_time

from logilab.common.fileutils import lines
from logilab.common.ureports import Table, TextWriter

from cubicweb import repoapi
from cubicweb.server.repository import Repository

TB_LOCK = threading.Lock()


class QueryExecutor:
    def __init__(self, cnx, times, queries, reporter=None):
        self._cnx = cnx
        self._times = times
        self._queries = queries
        self._reporter = reporter

    def run(self):
        with self._cnx as cnx:
            times = self._times
            while times:
                for index, query in enumerate(self._queries):
                    start = process_time()
                    try:
                        cnx.execute(query)
                    except Exception:
                        TB_LOCK.acquire()
                        traceback.print_exc()
                        TB_LOCK.release()
                        return
                    if self._reporter is not None:
                        self._reporter.add_proftime(process_time() - start, index)
                times -= 1


def usage(status=0):
    """print usage string and exit"""
    print(__doc__ % basename(sys.argv[0]))
    sys.exit(status)


class ProfileReporter:
    """a profile reporter gathers all profile informations from several
    threads and can write a report that summarizes all profile informations
    """

    profiler_lock = threading.Lock()

    def __init__(self, queries):
        self._queries = tuple(queries)
        self._profile_results = [(0.0, 0)] * len(self._queries)
        # self._table_report = Table(3, rheaders = True)
        len_max = max([len(query) for query in self._queries]) + 5
        self._query_fmt = "%%%ds" % len_max

    def add_proftime(self, elapsed_time, query_index):
        """add a new time measure for query"""
        ProfileReporter.profiler_lock.acquire()
        cumul_time, times = self._profile_results[query_index]
        cumul_time += elapsed_time
        times += 1.0
        self._profile_results[query_index] = (cumul_time, times)
        ProfileReporter.profiler_lock.release()

    def dump_report(self, output=sys.stdout):
        """dump report in 'output'"""
        table_elems = ["RQL Query", "Times", "Avg Time"]
        total_time = 0.0
        for query, (cumul_time, times) in zip(self._queries, self._profile_results):
            avg_time = cumul_time / float(times)
            table_elems += [str(query), f"{times:f}", f"{avg_time:f}"]
            total_time += cumul_time
        table_elems.append("Total time :")
        table_elems.append(str(total_time))
        table_elems.append(" ")
        table_layout = Table(3, rheaders=True, children=table_elems)
        TextWriter().format(table_layout, output)
        # output.write('\n'.join(tmp_output))


def run(args):
    """run the command line tool"""
    try:
        opts, args = getopt.getopt(
            args,
            "hn:t:u:p:P:o:",
            [
                "help",
                "user=",
                "password=",
                "nb-times=",
                "nb-threads=",
                "profile",
                "report-output=",
            ],
        )
    except Exception as ex:
        print(ex)
        usage(1)
    repeat = 100
    threads = 1
    user = os.environ.get("USER", os.environ.get("LOGNAME"))
    password = None
    report_output = sys.stdout
    prof_file = None
    for opt, val in opts:
        if opt in ("-h", "--help"):
            usage()
        if opt in ("-u", "--user"):
            user = val
        elif opt in ("-p", "--password"):
            password = val
        elif opt in ("-n", "--nb-times"):
            repeat = int(val)
        elif opt in ("-t", "--nb-threads"):
            threads = int(val)
        elif opt in ("-P", "--profile"):
            prof_file = val
        elif opt in ("-o", "--report-output"):
            report_output = open(val, "w")
    if len(args) != 2:
        usage(1)
    queries = [query for query in lines(args[1]) if not query.startswith("#")]
    if user is None:
        user = input("login: ")
    if password is None:
        password = getpass("password: ")
    from cubicweb.cwconfig import instance_configuration

    config = instance_configuration(args[0])
    # get local access to the repository
    print("Creating repo", prof_file)
    repo = Repository(config, prof_file)
    repo.bootstrap()
    cnx = repoapi.connect(repo, user, password=password)
    reporter = ProfileReporter(queries)
    if threads > 1:
        executors = []
        while threads:
            qe = QueryExecutor(cnx, repeat, queries, reporter=reporter)
            executors.append(qe)
            thread = threading.Thread(target=qe.run)
            qe.thread = thread
            thread.start()
            threads -= 1
        for qe in executors:
            qe.thread.join()
    else:
        QueryExecutor(cnx, repeat, queries, reporter=reporter).run()
    reporter.dump_report(report_output)


if __name__ == "__main__":
    run(sys.argv[1:])



# copyright 2003 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""classes to define schemas for CubicWeb"""

import re
from functools import wraps
from hashlib import md5
from logging import getLogger
from os.path import join

import yams
from logilab.common.decorators import cached, clear_cache, monkeypatch, cachedproperty
from logilab.common.graph import get_cycles
from logilab.common.logging_ext import set_log_methods
from logilab.common.textutils import splitstrip
from logilab.common.modutils import cleanup_sys_modules
from rql import parse, nodes, RQLSyntaxError, TypeResolverException
from rql.analyze import ETypeResolver
from yams import BadSchemaDefinition, ValidationError, buildobjs as ybo
from yams.buildobjs import _add_relation as yams_add_relation
from yams.constraints import (
    BaseConstraint,
    FormatConstraint,
    cstr_json_dumps,
    cstr_json_loads,
)
from yams.reader import (
    CONSTRAINTS,
    SchemaLoader,
    fill_schema_from_namespace,
)
from yams.schema import (
    Schema,
    ERSchema,
    EntitySchema,
    RelationSchema,
    RelationDefinitionSchema,
    PermissionMixIn,
    role_name,
)

import cubicweb
from cubicweb import Unauthorized, _
from cubicweb import server

PURE_VIRTUAL_RTYPES = set(
    (
        "identity",
        "has_text",
    )
)
VIRTUAL_RTYPES = set(
    (
        "eid",
        "identity",
        "has_text",
    )
)

# set of meta-relations available for every entity types
META_RTYPES = set(
    (
        "owned_by",
        "created_by",
        "is",
        "is_instance_of",
        "identity",
        "eid",
        "creation_date",
        "cw_source",
        "modification_date",
        "has_text",
        "cwuri",
    )
)
WORKFLOW_RTYPES = set(("custom_workflow", "in_state", "wf_info_for"))
WORKFLOW_DEF_RTYPES = set(
    (
        "workflow_of",
        "state_of",
        "transition_of",
        "initial_state",
        "default_workflow",
        "allowed_transition",
        "destination_state",
        "from_state",
        "to_state",
        "condition",
        "subworkflow",
        "subworkflow_state",
        "subworkflow_exit",
        "by_transition",
    )
)
SYSTEM_RTYPES = (
    set(
        (
            "in_group",
            "require_group",
            # cwproperty
            "for_user",
            "cw_import_of",
            "cw_host_config_of",
        )
    )
    | WORKFLOW_RTYPES
)
NO_I18NCONTEXT = META_RTYPES | WORKFLOW_RTYPES

SKIP_COMPOSITE_RELS = [("cw_source", "subject")]

# set of entity and relation types used to build the schema
SCHEMA_TYPES = set(
    (
        "CWEType",
        "CWRType",
        "CWComputedRType",
        "CWAttribute",
        "CWRelation",
        "CWConstraint",
        "CWConstraintType",
        "CWUniqueTogetherConstraint",
        "RQLExpression",
        "specializes",
        "relation_type",
        "from_entity",
        "to_entity",
        "constrained_by",
        "cstrtype",
        "constraint_of",
        "relations",
        "read_permission",
        "add_permission",
        "delete_permission",
        "update_permission",
    )
)

WORKFLOW_TYPES = set(
    (
        "Transition",
        "State",
        "TrInfo",
        "Workflow",
        "WorkflowTransition",
        "BaseTransition",
        "SubWorkflowExitPoint",
    )
)

INTERNAL_TYPES = set(
    (
        "CWProperty",
        "ExternalUri",
        "CWDataImport",
        "CWSource",
        "CWSourceHostConfig",
        "CWSession",
    )
)

UNIQUE_CONSTRAINTS = (
    "SizeConstraint",
    "FormatConstraint",
    "StaticVocabularyConstraint",
    "RQLVocabularyConstraint",
)

_LOGGER = getLogger("cubicweb.schemaloader")

# entity and relation schema created from serialized schema have an eid
ybo.ETYPE_PROPERTIES += ("eid",)
ybo.RTYPE_PROPERTIES += ("eid",)


def build_schema_from_namespace(items):
    schema = CubicWebSchema("noname")
    fill_schema_from_namespace(schema, items, register_base_types=False)
    return schema


# Bases for manipulating RQL in schema #########################################


def guess_rrqlexpr_mainvars(expression):
    defined = set(split_expression(expression))
    mainvars = set()
    if "S" in defined:
        mainvars.add("S")
    if "O" in defined:
        mainvars.add("O")
    if "U" in defined:
        mainvars.add("U")
    if not mainvars:
        raise BadSchemaDefinition(
            f"unable to guess selection variables in {expression!r}"
        )
    return mainvars


def split_expression(rqlstring):
    for expr in rqlstring.split(","):
        for noparen1 in expr.split("("):
            for noparen2 in noparen1.split(")"):
                for word in noparen2.split():
                    yield word


def normalize_expression(rqlstring):
    """normalize an rql expression to ease schema synchronization (avoid
    suppressing and reinserting an expression if only a space has been
    added/removed for instance)
    """
    union = parse(f"Any 1 WHERE {rqlstring}").as_string()
    return union.split(" WHERE ", 1)[1]


def _check_valid_formula(rdef, formula_rqlst):
    """Check the formula is a valid RQL query with some restriction (no union,
    single selected node, etc.), raise BadSchemaDefinition if not
    """
    if len(formula_rqlst.children) != 1:
        raise BadSchemaDefinition(
            "computed attribute %(attr)s on %(etype)s: "
            "can not use UNION in formula %(form)r"
            % {"attr": rdef.rtype, "etype": rdef.subject.type, "form": rdef.formula}
        )
    select = formula_rqlst.children[0]
    if len(select.selection) != 1:
        raise BadSchemaDefinition(
            "computed attribute %(attr)s on %(etype)s: "
            "can only select one term in formula %(form)r"
            % {"attr": rdef.rtype, "etype": rdef.subject.type, "form": rdef.formula}
        )
    term = select.selection[0]
    types = set(term.get_type(sol) for sol in select.solutions)
    if len(types) != 1:
        raise BadSchemaDefinition(
            "computed attribute %(attr)s on %(etype)s: "
            "multiple possible types (%(types)s) for formula %(form)r"
            % {
                "attr": rdef.rtype,
                "etype": rdef.subject.type,
                "types": list(types),
                "form": rdef.formula,
            }
        )
    computed_type = types.pop()
    expected_type = rdef.object.type
    if computed_type != expected_type:
        raise BadSchemaDefinition(
            "computed attribute %(attr)s on %(etype)s: "
            "computed attribute type (%(comp_type)s) mismatch with "
            "specified type (%(attr_type)s)"
            % {
                "attr": rdef.relation_type,
                "etype": rdef.subject.type,
                "comp_type": computed_type,
                "attr_type": expected_type,
            }
        )


class RQLExpression:
    """Base class for RQL expression used in schema (constraints and
    permissions)
    """

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None
    # to be defined in concrete classes
    predefined_variables = None

    # Internal cache for parsed expressions
    _rql_cache = {}

    @classmethod
    def _cached_parse(cls, rql):
        try:
            return cls._rql_cache[rql]
        except KeyError:
            cls._rql_cache[rql] = parse(rql, print_errors=False).children[0]
            return cls._rql_cache[rql]

    def __init__(self, expression, mainvars, eid):
        """
        :type mainvars: sequence of RQL variables' names. Can be provided as a
                        comma separated string.
        :param mainvars: names of the variables being selected.

        """
        self.eid = eid  # eid of the entity representing this rql expression
        assert mainvars, f"bad mainvars {mainvars}"
        if isinstance(mainvars, str):
            mainvars = set(splitstrip(mainvars))
        elif not isinstance(mainvars, set):
            mainvars = set(mainvars)
        self.mainvars = mainvars
        self.expression = normalize_expression(expression)
        try:
            # syntax tree used by read security (inserted in queries when necessary)
            self.snippet_rqlst = self._cached_parse(self.minimal_rql)
        except RQLSyntaxError:
            raise RQLSyntaxError(expression)
        for mainvar in mainvars:
            if len(self.snippet_rqlst.defined_vars[mainvar].references()) < 2:
                _LOGGER.warning(
                    "You did not use the %s variable in your RQL " "expression %s",
                    mainvar,
                    self,
                )
        # graph of links between variables, used by rql rewriter
        self.vargraph = vargraph(self.snippet_rqlst)
        # useful for some instrumentation, e.g. localperms permcheck command
        self.package = ybo.PACKAGE

    def __str__(self):
        return self.rqlst.as_string()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.expression})"

    def __lt__(self, other):
        if hasattr(other, "expression"):
            return self.expression < other.expression
        return True

    def __eq__(self, other):
        if hasattr(other, "expression"):
            return self.expression == other.expression
        return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.expression)

    def __deepcopy__(self, memo):
        return self.__class__(self.expression, self.mainvars)

    def __getstate__(self):
        return (self.expression, self.mainvars)

    def __setstate__(self, state):
        self.__init__(*state)

    @cachedproperty
    def rqlst(self):
        # Don't use _cached_parse here because the rqlst is modified
        select = parse(self.minimal_rql, print_errors=False).children[0]
        defined = set(split_expression(self.expression))
        for varname in self.predefined_variables:
            if varname in defined:
                select.add_eid_restriction(
                    select.get_variable(varname), varname.lower(), "Substitute"
                )
        return select

    # permission rql expression specific stuff #################################

    @cached
    def transform_has_permission(self):
        found = None
        rqlst = self.rqlst
        for var in rqlst.defined_vars.values():
            for varref in var.references():
                rel = varref.relation()
                if rel is None:
                    continue
                try:
                    prefix, action, suffix = rel.r_type.split("_")
                except ValueError:
                    continue
                if (
                    prefix != "has"
                    or suffix != "permission"
                    or action not in ("add", "delete", "update", "read")
                ):
                    continue
                if found is None:
                    found = []
                    rqlst.save_state()
                assert rel.children[0].name == "U"
                objvar = rel.children[1].children[0].variable
                rqlst.remove_node(rel)
                selected = [v.name for v in rqlst.get_selected_variables()]
                if objvar.name not in selected:
                    colindex = len(selected)
                    rqlst.add_selected(objvar)
                else:
                    colindex = selected.index(objvar.name)
                found.append((action, colindex))
                # remove U eid %(u)s if U is not used in any other relation
                uvrefs = rqlst.defined_vars["U"].references()
                if len(uvrefs) == 1:
                    rqlst.remove_node(uvrefs[0].relation())
        if found is not None:
            rql = rqlst.as_string()
            if len(rqlst.selection) == 1 and isinstance(rqlst.where, nodes.Relation):
                # only "Any X WHERE X eid %(x)s" remaining, no need to execute the rql
                keyarg = rqlst.selection[0].name.lower()
            else:
                keyarg = None
            rqlst.recover()
            return rql, found, keyarg
        return rqlst.as_string(), None, None

    def _check(self, _cw, **kwargs):
        """return True if the rql expression is matching the given relation
        between fromeid and toeid

        _cw may be a request or a server side transaction
        """
        creating = kwargs.get("creating")
        if not creating and self.eid is not None:
            key = (self.eid, tuple(sorted(kwargs.items())))
            try:
                return _cw.local_perm_cache[key]
            except KeyError:
                pass
        rql, has_perm_defs, keyarg = self.transform_has_permission()
        # when creating an entity, expression related to X satisfied
        if creating and "X" in self.snippet_rqlst.defined_vars:
            return True
        if keyarg is None:
            kwargs.setdefault("u", _cw.user.eid)
            try:
                # ensure security is disabled
                with getattr(_cw, "cnx", _cw).security_enabled(read=False):
                    rset = _cw.execute(rql, kwargs, build_descr=True)
            except NotImplementedError:
                self.critical("cant check rql expression, unsupported rql %s", rql)
                if self.eid is not None:
                    _cw.local_perm_cache[key] = False
                return False
            except TypeResolverException as ex:
                # some expression may not be resolvable with current kwargs
                # (type conflict)
                self.warning("%s: %s", rql, str(ex))
                if self.eid is not None:
                    _cw.local_perm_cache[key] = False
                return False
            except Unauthorized as ex:
                self.debug("unauthorized %s: %s", rql, str(ex))
                if self.eid is not None:
                    _cw.local_perm_cache[key] = False
                return False
        else:
            rset = _cw.eid_rset(kwargs[keyarg])
        # if no special has_*_permission relation in the rql expression, just
        # check the result set contains something
        if has_perm_defs is None:
            if rset:
                if self.eid is not None:
                    _cw.local_perm_cache[key] = True
                return True
        elif rset:
            # check every special has_*_permission relation is satisfied
            get_eschema = _cw.vreg.schema.entity_schema_for
            try:
                for eaction, col in has_perm_defs:
                    for i in range(len(rset)):
                        eschema = get_eschema(rset.description[i][col])
                        eschema.check_perm(_cw, eaction, eid=rset[i][col])
                if self.eid is not None:
                    _cw.local_perm_cache[key] = True
                return True
            except Unauthorized:
                pass
        if self.eid is not None:
            _cw.local_perm_cache[key] = False
        return False

    @property
    def minimal_rql(self):
        return f"Any {','.join(sorted(self.mainvars))} WHERE {self.expression}"


# rql expressions for use in permission definition #############################


class ERQLExpression(RQLExpression):
    predefined_variables = "XU"

    def __init__(self, expression, mainvars=None, eid=None):
        RQLExpression.__init__(self, expression, mainvars or "X", eid)

    def check(self, _cw, eid=None, creating=False, **kwargs):
        if "X" in self.snippet_rqlst.defined_vars:
            if eid is None:
                if creating:
                    return self._check(_cw, creating=True, **kwargs)
                return False
            assert not creating
            return self._check(_cw, x=eid, **kwargs)
        return self._check(_cw, **kwargs)


class CubicWebRelationDefinitionSchema(RelationDefinitionSchema):
    def constraint_by_eid(self, eid):
        for cstr in self.constraints:
            if cstr.eid == eid:
                return cstr
        raise ValueError("No constraint with eid %d" % eid)

    def rql_expression(self, expression, mainvars=None, eid=None):
        """rql expression factory"""
        if self.relation_type.final:
            return ERQLExpression(expression, mainvars, eid)
        return RRQLExpression(expression, mainvars, eid)

    def check_permission_definitions(self):
        super(CubicWebRelationDefinitionSchema, self).check_permission_definitions()
        for action, groups in self.permissions.items():
            for group_or_rqlexpr in groups:
                if action == "read" and isinstance(group_or_rqlexpr, RQLExpression):
                    msg = "can't use rql expression for read permission of %s"
                    raise BadSchemaDefinition(msg % self)
                if self.final and isinstance(group_or_rqlexpr, RRQLExpression):
                    msg = "can't use RRQLExpression on %s, use an ERQLExpression"
                    raise BadSchemaDefinition(msg % self)
                if not self.final and isinstance(group_or_rqlexpr, ERQLExpression):
                    msg = "can't use ERQLExpression on %s, use a RRQLExpression"
                    raise BadSchemaDefinition(msg % self)


def vargraph(rqlst):
    """builds an adjacency graph of variables from the rql syntax tree, e.g:
    Any O,S WHERE T subworkflow_exit S, T subworkflow WF, O state_of WF
    => {'WF': ['O', 'T'], 'S': ['T'], 'T': ['WF', 'S'], 'O': ['WF']}
    """
    vargraph = {}
    for relation in rqlst.get_nodes(nodes.Relation):
        try:
            rhsvarname = relation.children[1].children[0].variable.name
            lhsvarname = relation.children[0].name
        except AttributeError:
            pass
        else:
            vargraph.setdefault(lhsvarname, []).append(rhsvarname)
            vargraph.setdefault(rhsvarname, []).append(lhsvarname)
    return vargraph


class GeneratedConstraint:
    def __init__(self, rqlst, mainvars):
        self.snippet_rqlst = rqlst
        self.mainvars = mainvars
        self.vargraph = vargraph(rqlst)


class RRQLExpression(RQLExpression):
    predefined_variables = "SOU"

    def __init__(self, expression, mainvars=None, eid=None):
        if mainvars is None:
            mainvars = guess_rrqlexpr_mainvars(expression)
        RQLExpression.__init__(self, expression, mainvars, eid)

    def check(self, _cw, fromeid=None, toeid=None):
        kwargs = {}
        if "S" in self.snippet_rqlst.defined_vars:
            if fromeid is None:
                return False
            kwargs["s"] = fromeid
        if "O" in self.snippet_rqlst.defined_vars:
            if toeid is None:
                return False
            kwargs["o"] = toeid
        return self._check(_cw, **kwargs)


# In yams, default 'update' perm for attributes granted to managers and owners.
# Within cw, we want to default to users who may edit the entity holding the
# attribute.
# These default permissions won't be checked by the security hooks:
# since they delegate checking to the entity, we can skip actual checks.
ybo.DEFAULT_ATTRPERMS["update"] = (
    "managers",
    ERQLExpression("U has_update_permission X"),
)
ybo.DEFAULT_ATTRPERMS["add"] = ("managers", ERQLExpression("U has_add_permission X"))

# we don't want 'add' or 'delete' permissions on computed relation types
# (they're hardcoded to '()' on computed relation definitions)
if "add" in yams.DEFAULT_COMPUTED_RELPERMS:
    del yams.DEFAULT_COMPUTED_RELPERMS["add"]
if "delete" in yams.DEFAULT_COMPUTED_RELPERMS:
    del yams.DEFAULT_COMPUTED_RELPERMS["delete"]


PUB_SYSTEM_ENTITY_PERMS = {
    "read": (
        "managers",
        "users",
        "guests",
    ),
    "add": ("managers",),
    "delete": ("managers",),
    "update": ("managers",),
}
PUB_SYSTEM_REL_PERMS = {
    "read": (
        "managers",
        "users",
        "guests",
    ),
    "add": ("managers",),
    "delete": ("managers",),
}
PUB_SYSTEM_ATTR_PERMS = {
    "read": (
        "managers",
        "users",
        "guests",
    ),
    "add": ("managers",),
    "update": ("managers",),
}
RO_REL_PERMS = {
    "read": (
        "managers",
        "users",
        "guests",
    ),
    "add": (),
    "delete": (),
}
RO_ATTR_PERMS = {
    "read": (
        "managers",
        "users",
        "guests",
    ),
    "add": ybo.DEFAULT_ATTRPERMS["add"],
    "update": (),
}


# XXX same algorithm as in reorder_cubes and probably other place,
# may probably extract a generic function
def order_eschemas(eschemas):
    """return entity schemas ordered such that entity types which specializes an
    other one appears after that one
    """
    graph = {}
    for eschema in eschemas:
        if eschema.specializes():
            graph[eschema] = set((eschema.specializes(),))
        else:
            graph[eschema] = set()
    cycles = get_cycles(graph)
    if cycles:
        cycles = "\n".join(" -> ".join(cycle) for cycle in cycles)
        raise Exception(f"cycles in entity schema specialization: {cycles}")
    eschemas = []
    while graph:
        # sorted to get predictable results
        for eschema, deps in sorted(graph.items()):
            if not deps:
                eschemas.append(eschema)
                del graph[eschema]
                for deps in graph.values():
                    try:
                        deps.remove(eschema)
                    except KeyError:
                        continue
    return eschemas


def display_name(req, key, form="", context=None):
    """return a internationalized string for the key (schema entity or relation
    name) in a given form
    """
    assert form in ("", "plural", "subject", "object")
    if form == "subject":
        form = ""
    if form:
        key = key + "_" + form
    # ensure unicode
    if context is not None:
        return req.pgettext(context, key)
    else:
        return req._(key)


def _override_method(cls, method_name=None, pass_original=False):
    """Override (or set) a method on `cls`."""

    def decorator(func):
        name = method_name or func.__name__
        orig = None
        if pass_original:
            orig = getattr(cls, name)

        @wraps(func)
        def wrapper(*args, **kwargs):
            if orig is not None:
                kwargs["_orig"] = orig
            return func(*args, **kwargs)

        setattr(cls, name, wrapper)

    return decorator


# Schema objects definition ###################################################


@_override_method(ERSchema, "display_name")
def ERSchema_display_name(self, req, form="", context=None):
    """return a internationalized string for the entity/relation type name in
    a given form
    """
    return display_name(req, self.type, form, context)


@_override_method(PermissionMixIn)
@cached
def get_groups(self, action):
    """return the groups authorized to perform <action> on entities of
    this type

    :type action: str
    :param action: the name of a permission

    :rtype: tuple
    :return: names of the groups with the given permission
    """
    assert action in self.ACTIONS, action
    try:
        return frozenset(g for g in self.permissions[action] if isinstance(g, str))
    except KeyError:
        return ()


@_override_method(PermissionMixIn)
@cached
def get_rqlexprs(self, action):
    """return the rql expressions representing queries to check the user is allowed
    to perform <action> on entities of this type

    :type action: str
    :param action: the name of a permission

    :rtype: tuple
    :return: the rql expressions with the given permission
    """
    assert action in self.ACTIONS, action
    try:
        return tuple(g for g in self.permissions[action] if not isinstance(g, str))
    except KeyError:
        return ()


@_override_method(PermissionMixIn, pass_original=True)
def set_action_permissions(self, action, permissions, _orig):
    """set the groups and rql expressions allowing to perform <action> on
    entities of this type

    :type action: str
    :param action: the name of a permission

    :type permissions: tuple
    :param permissions: the groups and rql expressions allowing the given action
    """
    _orig(self, action, tuple(permissions))
    clear_cache(self, "get_rqlexprs")
    clear_cache(self, "get_groups")


@_override_method(PermissionMixIn)
def has_local_role(self, action):
    """return true if the action *may* be granted locally (i.e. either rql
    expressions or the owners group are used in security definition)

    XXX this method is only there since we don't know well how to deal with
    'add' action checking. Also find a better name would be nice.
    """
    assert action in self.ACTIONS, action
    if self.get_rqlexprs(action):
        return True
    if action in ("update", "delete"):
        return "owners" in self.get_groups(action)
    return False


@_override_method(PermissionMixIn)
def may_have_permission(self, action, req):
    if action != "read" and not (
        self.has_local_role("read") or self.has_perm(req, "read")
    ):
        return False
    return self.has_local_role(action) or self.has_perm(req, action)


@_override_method(PermissionMixIn)
def has_perm(self, _cw, action, **kwargs):
    """return true if the action is granted globally or locally"""
    try:
        self.check_perm(_cw, action, **kwargs)
        return True
    except Unauthorized:
        return False


@_override_method(PermissionMixIn)
def check_perm(self, _cw, action, **kwargs):
    # NB: _cw may be a server transaction or a request object.
    #
    # check user is in an allowed group, if so that's enough internal
    # transactions should always stop there
    DBG = False
    if server.DEBUG & server.DBG_SEC:
        if action in server._SECURITY_CAPS:
            _self_str = str(self)
            if server._SECURITY_ITEMS:
                if any(item in _self_str for item in server._SECURITY_ITEMS):
                    DBG = True
            else:
                DBG = True
    groups = self.get_groups(action)
    if _cw.user.matching_groups(groups):
        if DBG:
            print(f"check_perm: {action!r} {_self_str!r}: user matches {groups}")
        return
    # if 'owners' in allowed groups, check if the user actually owns this
    # object, if so that's enough
    #
    # NB: give _cw to user.owns since user is not be bound to a transaction on
    # the repository side
    if "owners" in groups and (
        kwargs.get("creating") or ("eid" in kwargs and _cw.user.owns(kwargs["eid"]))
    ):
        if DBG:
            print(
                f"check_perm: {action!r} {_self_str!r}: user is owner or creation time"
            )
        return
    # else if there is some rql expressions, check them
    if DBG:
        print(
            "check_perm: %r %r %s"
            % (
                action,
                _self_str,
                [
                    (rqlexpr, kwargs, rqlexpr.check(_cw, **kwargs))
                    for rqlexpr in self.get_rqlexprs(action)
                ],
            )
        )
    if any(rqlexpr.check(_cw, **kwargs) for rqlexpr in self.get_rqlexprs(action)):
        return
    raise Unauthorized(action, str(self))


CubicWebRelationDefinitionSchema._RPROPERTIES["eid"] = None
# remember rproperties defined at this point. Others will have to be serialized in
# CWAttribute.extra_props
KNOWN_RPROPERTIES = CubicWebRelationDefinitionSchema.ALL_PROPERTIES()


class CubicWebEntitySchema(EntitySchema):
    """a entity has a type, a set of subject and or object relations
    the entity schema defines the possible relations for a given type and some
    constraints on those relations
    """

    def __init__(self, schema=None, edef=None, eid=None, **kwargs):
        super(CubicWebEntitySchema, self).__init__(schema, edef, **kwargs)
        if eid is None and edef is not None:
            eid = getattr(edef, "eid", None)
        self.eid = eid

    def targets(self, role):
        assert role in ("subject", "object")
        if role == "subject":
            return self.subject_relations.values()
        return self.object_relations.values()

    @cachedproperty
    def composite_rdef_roles(self):
        """Return all relation definitions that define the current entity
        type as a composite.
        """
        rdef_roles = []
        for role in ("subject", "object"):
            for rschema in self.targets(role):
                if rschema.final:
                    continue
                for rdef in rschema.relation_definitions.values():
                    if (role == "subject" and rdef.subject == self) or (
                        role == "object" and rdef.object == self
                    ):
                        crole = rdef.composite
                        if crole == role:
                            rdef_roles.append((rdef, role))
        return rdef_roles

    @cachedproperty
    def is_composite(self):
        return bool(len(self.composite_rdef_roles))

    def check_permission_definitions(self):
        super(CubicWebEntitySchema, self).check_permission_definitions()
        for groups in self.permissions.values():
            for group_or_rqlexpr in groups:
                if isinstance(group_or_rqlexpr, RRQLExpression):
                    msg = "can't use RRQLExpression on %s, use an ERQLExpression"
                    raise BadSchemaDefinition(msg % self.type)

    def is_subobject(self, strict=False, skiprels=None):
        if skiprels is None:
            skiprels = SKIP_COMPOSITE_RELS
        else:
            skiprels += SKIP_COMPOSITE_RELS
        return super(CubicWebEntitySchema, self).is_subobject(
            strict, skip_relations=skiprels
        )

    def attribute_definitions(self):
        """return an iterator on attribute definitions

        attribute relations are a subset of subject relations where the
        object's type is a final entity

        an attribute definition is a 2-uple :
        * name of the relation
        * schema of the destination entity type
        """
        iter = super(CubicWebEntitySchema, self).attribute_definitions()
        for rschema, attrschema in iter:
            if rschema.type == "has_text":
                continue
            yield rschema, attrschema

    def main_attribute(self):
        """convenience method that returns the *main* (i.e. the first non meta)
        attribute defined in the entity schema
        """
        for rschema, __ in self.attribute_definitions():
            if not (rschema in META_RTYPES or self.is_metadata(rschema)):
                return rschema

    def add_subject_relation(self, rschema):
        """register the relation schema as possible subject relation"""
        super(CubicWebEntitySchema, self).add_subject_relation(rschema)
        if rschema.final:
            if self.relation_definition(rschema).get("fulltextindexed"):
                self._update_has_text()
        elif rschema.fulltext_container:
            self._update_has_text()

    def add_object_relation(self, rschema):
        """register the relation schema as possible object relation"""
        super(CubicWebEntitySchema, self).add_object_relation(rschema)
        if rschema.fulltext_container:
            self._update_has_text()

    def del_subject_relation(self, rtype):
        super(CubicWebEntitySchema, self).del_subject_relation(rtype)
        if "has_text" in self.subject_relations:
            self._update_has_text(deletion=True)

    def del_object_relation(self, rtype):
        super(CubicWebEntitySchema, self).del_object_relation(rtype)
        if "has_text" in self.subject_relations:
            self._update_has_text(deletion=True)

    def _update_has_text(self, deletion=False):
        may_need_has_text, has_has_text = False, False
        need_has_text = None
        for rschema in list(self.subject_relations.values()):
            if rschema.final:
                if rschema == "has_text":
                    has_has_text = True
                elif self.relation_definition(rschema).get("fulltextindexed"):
                    may_need_has_text = True
            elif rschema.fulltext_container:
                if rschema.fulltext_container == "subject":
                    may_need_has_text = True
                else:
                    need_has_text = False
        for rschema in list(self.object_relations.values()):
            if rschema.fulltext_container:
                if rschema.fulltext_container == "object":
                    may_need_has_text = True
                else:
                    need_has_text = False
        if need_has_text is None:
            need_has_text = may_need_has_text
        if need_has_text and not has_has_text and not deletion:
            rdef = ybo.RelationDefinition(
                self.type, "has_text", "String", __permissions__=RO_ATTR_PERMS
            )
            self.schema.add_relation_def(rdef)
        elif not need_has_text and has_has_text:
            # use rschema.del_relation_def and not schema.del_relation_def to
            # avoid deleting the relation type accidentally...
            self.schema["has_text"].del_relation_def(self, self.schema["String"])

    def schema_entity(self):  # XXX @property for consistency with meta
        """return True if this entity type is used to build the schema"""
        return self.type in SCHEMA_TYPES

    def rql_expression(self, expression, mainvars=None, eid=None):
        """rql expression factory"""
        return ERQLExpression(expression, mainvars, eid)


class CubicWebRelationSchema(PermissionMixIn, RelationSchema):
    permissions = {}
    ACTIONS = ()
    relation_definition_class = CubicWebRelationDefinitionSchema

    # retro-compatibility for yams <= 0.45.*
    # this situation is not handled by our deprecation tools
    rdef_class = relation_definition_class

    def __init__(self, schema=None, rdef=None, eid=None, **kwargs):
        if rdef is not None:
            # if this relation is inlined
            self.inlined = rdef.inlined
        super(CubicWebRelationSchema, self).__init__(schema, rdef, **kwargs)
        if eid is None and rdef is not None:
            eid = getattr(rdef, "eid", None)
        self.eid = eid

    def init_computed_relation(self, rdef):
        # retro-compat for older version of yams (0.45.x and before) where
        # init_computed_relation has been renamed to _init_computed_relation
        # this situation is not handled by our deprecation tools
        self._init_computed_relation(rdef)

    def _init_computed_relation(self, rdef):
        self.ACTIONS = ("read",)

        # retro-compat for older version of yams (0.45.x and before) where
        # init_computed_relation has been renamed to _init_computed_relation
        # this situation is not handled by our deprecation tools
        if hasattr(super(CubicWebRelationSchema, self), "_init_computed_relation"):
            super(CubicWebRelationSchema, self)._init_computed_relation(rdef)
        else:
            super(CubicWebRelationSchema, self).init_computed_relation(rdef)

    def advertise_new_add_permission(self):
        pass

    def check_permission_definitions(self):
        RelationSchema.check_permission_definitions(self)
        PermissionMixIn.check_permission_definitions(self)

    @property
    def meta(self):
        return self.type in META_RTYPES

    def schema_relation(self):  # XXX @property for consistency with meta
        """return True if this relation type is used to build the schema"""
        return self.type in SCHEMA_TYPES

    def may_have_permission(self, action, req, eschema=None, role=None):
        if eschema is not None:
            for tschema in self.targets(eschema, role):
                rdef = self.role_relation_definition(eschema, tschema, role)
                if rdef.may_have_permission(action, req):
                    return True
        else:
            for rdef in self.relation_definitions.values():
                if rdef.may_have_permission(action, req):
                    return True
        return False

    def has_perm(self, _cw, action, **kwargs):
        """return true if the action is granted globally or locally"""
        if self.final:
            assert not ("fromeid" in kwargs or "toeid" in kwargs), kwargs
            assert action in ("read", "update")
            if "eid" in kwargs:
                subjtype = _cw.entity_type(kwargs["eid"])
            else:
                subjtype = objtype = None
        else:
            assert "eid" not in kwargs, kwargs
            assert action in ("read", "add", "delete")
            if "fromeid" in kwargs:
                subjtype = _cw.entity_type(kwargs["fromeid"])
            elif "frometype" in kwargs:
                subjtype = kwargs.pop("frometype")
            else:
                subjtype = None
            if "toeid" in kwargs:
                objtype = _cw.entity_type(kwargs["toeid"])
            elif "toetype" in kwargs:
                objtype = kwargs.pop("toetype")
            else:
                objtype = None
        if objtype and subjtype:
            return self.relation_definition(subjtype, objtype).has_perm(
                _cw, action, **kwargs
            )
        elif subjtype:
            for tschema in self.targets(subjtype, "subject"):
                rdef = self.relation_definition(subjtype, tschema)
                if not rdef.has_perm(_cw, action, **kwargs):
                    return False
        elif objtype:
            for tschema in self.targets(objtype, "object"):
                rdef = self.relation_definition(tschema, objtype)
                if not rdef.has_perm(_cw, action, **kwargs):
                    return False
        else:
            for rdef in self.relation_definitions.values():
                if not rdef.has_perm(_cw, action, **kwargs):
                    return False
        return True


class CubicWebSchema(Schema):
    """set of entities and relations schema defining the possible data sets
    used in an application

    :type name: str
    :ivar name: name of the schema, usually the instance identifier

    :type base: str
    :ivar base: path of the directory where the schema is defined
    """

    reading_from_database = False
    entity_class = CubicWebEntitySchema
    relation_class = CubicWebRelationSchema
    no_specialization_inference = ("identity",)

    def __init__(self, *args, **kwargs):
        self._eid_index = {}
        super(CubicWebSchema, self).__init__(*args, **kwargs)
        ybo.register_base_types(self)
        rschema = self.add_relation_type(ybo.RelationType("eid"))
        rschema.final = True
        rschema = self.add_relation_type(ybo.RelationType("has_text"))
        rschema.final = True
        rschema = self.add_relation_type(ybo.RelationType("identity"))
        rschema.final = False

    etype_name_re = r"[A-Z][A-Za-z0-9]*[a-z]+[A-Za-z0-9]*$"

    @cachedproperty
    def rules_rqlexpr_mapping(self):
        """Return a dictionary mapping rtype to RRQLExpression for computed
        relations.
        """
        rules = {}
        for rschema in self.iter_computed_relations():
            rules[rschema.type] = RRQLExpression(rschema.rule)
        return rules

    def add_entity_type(self, edef: yams.types.EntityType) -> CubicWebEntitySchema:
        edef.name = str(edef.name)
        if not re.match(self.etype_name_re, edef.name):
            raise BadSchemaDefinition(
                "%r is not a valid name for an entity type. It should start "
                "with an upper cased letter and be followed by at least a "
                "lower cased letter" % edef.name
            )
        eschema = super().add_entity_type(edef)
        if not eschema.final:
            # automatically add the eid relation to non final entity types
            rdef = ybo.RelationDefinition(
                eschema.type,
                "eid",
                "Int",
                cardinality="11",
                uid=True,
                __permissions__=RO_ATTR_PERMS,
            )
            self.add_relation_def(rdef)
            rdef = ybo.RelationDefinition(
                eschema.type, "identity", eschema.type, __permissions__=RO_REL_PERMS
            )
            self.add_relation_def(rdef)
        self._eid_index[eschema.eid] = eschema
        return eschema

    def add_relation_type(self, rdef):
        if not rdef.name.islower():
            raise BadSchemaDefinition(
                "%r is not a valid name for a relation type. It should be "
                "lower cased" % rdef.name
            )
        rdef.name = str(rdef.name)
        rschema = super(CubicWebSchema, self).add_relation_type(rdef)
        self._eid_index[rschema.eid] = rschema
        return rschema

    def add_relation_def(self, rdef):
        """build a part of a relation schema
        (i.e. add a relation between two specific entity's types)

        :type subject: str
        :param subject: entity's type that is subject of the relation

        :type rtype: str
        :param rtype: the relation's type (i.e. the name of the relation)

        :type obj: str
        :param obj: entity's type that is object of the relation

        :rtype: RelationSchema
        :param: the newly created or just completed relation schema
        """
        rdef.name = rdef.name.lower()
        rdefs = super(CubicWebSchema, self).add_relation_def(rdef)
        if rdefs:
            try:
                self._eid_index[rdef.eid] = rdefs
            except AttributeError:
                pass  # not a serialized schema
        return rdefs

    def del_relation_type(self, rtype):
        rschema = self.relation_schema_for(rtype)
        self._eid_index.pop(rschema.eid, None)
        super(CubicWebSchema, self).del_relation_type(rschema)

    def del_relation_def(self, subjtype, rtype, objtype):
        for k, v in self._eid_index.items():
            if not isinstance(v, RelationDefinitionSchema):
                continue
            if (
                v.subject == subjtype
                and v.relation_type == rtype
                and v.object == objtype
            ):
                del self._eid_index[k]
                break
        super(CubicWebSchema, self).del_relation_def(subjtype, rtype, objtype)

    def del_entity_type(self, etype):
        eschema = self.entity_schema_for(etype)
        self._eid_index.pop(eschema.eid, None)
        # deal with has_text first, else its automatic deletion (see above)
        # may trigger an error in ancestor's del_entity_type method
        if "has_text" in list(eschema.subject_relations.values()):
            self.del_relation_def(etype, "has_text", "String")
        super(CubicWebSchema, self).del_entity_type(etype)

    def schema_by_eid(self, eid):
        return self._eid_index[eid]

    def iter_computed_attributes(self):
        for relation in self.relations():
            for rdef in relation.relation_definitions.values():
                if rdef.final and rdef.formula is not None:
                    yield rdef

    def iter_computed_relations(self):
        for relation in self.relations():
            if relation.rule:
                yield relation

    def finalize(self):
        super(CubicWebSchema, self).finalize()
        self.finalize_computed_attributes()
        self.finalize_computed_relations()

    def finalize_computed_attributes(self):
        """Check computed attributes validity (if any), else raise
        `BadSchemaDefinition`
        """
        analyzer = ETypeResolver(self)
        for rdef in self.iter_computed_attributes():
            rqlst = parse(rdef.formula)
            select = rqlst.children[0]
            select.add_type_restriction(select.defined_vars["X"], str(rdef.subject))
            analyzer.visit(select)
            _check_valid_formula(rdef, rqlst)
            rdef.formula_select = select  # avoid later recomputation

    def finalize_computed_relations(self):
        """Build relation definitions for computed relations

        The subject and object types are infered using rql analyzer.
        """
        analyzer = ETypeResolver(self)
        for rschema in self.iter_computed_relations():
            # XXX rule is valid if both S and O are defined and not in an exists
            rqlexpr = RRQLExpression(rschema.rule)
            rqlst = rqlexpr.snippet_rqlst
            analyzer.visit(rqlst)
            couples = set((sol["S"], sol["O"]) for sol in rqlst.solutions)
            for subjtype, objtype in couples:
                if self[objtype].final:
                    raise BadSchemaDefinition("computed relations cannot be final")
                rdef = ybo.RelationDefinition(
                    subjtype,
                    rschema.type,
                    objtype,
                    __permissions__={
                        "add": (),
                        "delete": (),
                        "read": rschema.permissions["read"],
                    },
                )
                rdef.infered = True
                self.add_relation_def(rdef)

    def rebuild_infered_relations(self):
        super(CubicWebSchema, self).rebuild_infered_relations()
        self.finalize_computed_attributes()
        self.finalize_computed_relations()
        # remove @cachedproperty cache
        self.__dict__.pop("rules_rqlexpr_mapping", None)


# additional cw specific constraints ###########################################


def constraint_name_for(constraint, rdef):
    """Return a unique, size controlled, name for this constraint applied to given `rdef`.

    This name may be used as name for the constraint in the database.
    """
    return (
        "cstr"
        + md5(
            (
                rdef.subject.type
                + rdef.relation_type.type
                + constraint.type()
                + (constraint.serialize() or "")
            ).encode("ascii")
        ).hexdigest()
    )


class BaseRQLConstraint(RRQLExpression, BaseConstraint):
    """base class for rql constraints"""

    distinct_query = None

    def serialize(self):
        return cstr_json_dumps(
            {"mainvars": sorted(self.mainvars), "expression": self.expression}
        )

    @classmethod
    def deserialize(cls, value):
        try:
            d = cstr_json_loads(value)
            return cls(d["expression"], d["mainvars"])
        except ValueError:
            _, mainvars, expression = value.split(";", 2)
            return cls(expression, mainvars)

    def check(self, entity, rtype, value):
        """return true if the value satisfy the constraint, else false"""
        # implemented as a hook in the repository
        return 1

    def __str__(self):
        if self.distinct_query:
            selop = "Any"
        else:
            selop = "DISTINCT Any"
        return "%s(%s %s WHERE %s)" % (
            self.__class__.__name__,
            selop,
            ",".join(sorted(self.mainvars)),
            self.expression,
        )

    def __repr__(self):
        return "<%s @%#x>" % (self.__str__(), id(self))


class RQLVocabularyConstraint(BaseRQLConstraint):
    """the rql vocabulary constraint:

    limits the proposed values to a set of entities returned by an rql query,
    but this is not enforced at the repository level

    `expression` is an additional rql restriction that will be added to
    a predefined query, where the S and O variables respectively represent
    the subject and the object of the relation

    `mainvars` is a set of variables that should be used as selection variables
    (i.e. `'Any %s WHERE ...' % mainvars`). If not specified, an attempt will be
    made to guess it based on the variables used in the expression.
    """

    def repo_check(self, session, eidfrom, rtype, eidto):
        """raise ValidationError if the relation doesn't satisfy the constraint"""
        pass  # this is a vocabulary constraint, not enforced


class RepoEnforcedRQLConstraintMixIn:
    def __init__(self, expression, mainvars=None, msg=None):
        super(RepoEnforcedRQLConstraintMixIn, self).__init__(expression, mainvars)
        self.msg = msg

    def serialize(self):
        return cstr_json_dumps(
            {
                "mainvars": sorted(self.mainvars),
                "expression": self.expression,
                "msg": self.msg,
            }
        )

    @classmethod
    def deserialize(cls, value):
        try:
            d = cstr_json_loads(value)
            return cls(d["expression"], d["mainvars"], d["msg"])
        except ValueError:
            value, msg = value.split("\n", 1)
            _, mainvars, expression = value.split(";", 2)
            return cls(expression, mainvars, msg)

    def repo_check(self, session, eidfrom, rtype, eidto=None):
        """raise ValidationError if the relation doesn't satisfy the constraint"""
        if not self.match_condition(session, eidfrom, eidto):
            # XXX at this point if both or neither of S and O are in mainvar we
            # dunno if the validation error `occurred` on eidfrom or eidto (from
            # user interface point of view)
            #
            # possible enhancement: check entity being created, it's probably
            # the main eid unless this is a composite relation
            if eidto is None or "S" in self.mainvars or "O" not in self.mainvars:
                maineid = eidfrom
                qname = role_name(rtype, "subject")
            else:
                maineid = eidto
                qname = role_name(rtype, "object")
            if self.msg:
                msg = session._(self.msg)
            else:
                msg = f"{session._(self.type())} {self.expression} failed"
            raise ValidationError(maineid, {qname: msg})

    def exec_query(self, _cw, eidfrom, eidto):
        if eidto is None:
            # checking constraint for an attribute relation
            expression = "S eid %(s)s, " + self.expression
            args = {"s": eidfrom}
        else:
            expression = "S eid %(s)s, O eid %(o)s, " + self.expression
            args = {"s": eidfrom, "o": eidto}
        if "U" in self.snippet_rqlst.defined_vars:
            expression = "U eid %(u)s, " + expression
            args["u"] = _cw.user.eid
        rql = f"Any {','.join(sorted(self.mainvars))} WHERE {expression}"
        if self.distinct_query:
            rql = "DISTINCT " + rql
        return _cw.execute(rql, args, build_descr=False)


class RQLConstraint(RepoEnforcedRQLConstraintMixIn, BaseRQLConstraint):
    """the rql constraint is similar to the RQLVocabularyConstraint but
    are also enforced at the repository level
    """

    distinct_query = False

    def match_condition(self, session, eidfrom, eidto):
        return self.exec_query(session, eidfrom, eidto)


class RQLUniqueConstraint(RepoEnforcedRQLConstraintMixIn, BaseRQLConstraint):
    """the unique rql constraint check that the result of the query isn't
    greater than one.

    You *must* specify `mainvars` when instantiating the constraint since there
    is no way to guess it correctly (e.g. if using S,O or U the constraint will
    always be satisfied because we've to use a DISTINCT query).
    """

    # XXX turns mainvars into a required argument in __init__
    distinct_query = True

    def match_condition(self, session, eidfrom, eidto):
        return len(self.exec_query(session, eidfrom, eidto)) <= 1


# workflow extensions #########################################################


class workflowable_definition(ybo.metadefinition):
    """extends default EntityType's metaclass to add workflow relations
    (i.e. in_state, wf_info_for and custom_workflow). This is the default
    metaclass for WorkflowableEntityType.
    """

    def __new__(mcs, name, bases, classdict):
        abstract = classdict.pop("__abstract__", False)
        cls = super(workflowable_definition, mcs).__new__(mcs, name, bases, classdict)
        if not abstract:
            make_workflowable(cls)
        return cls


class WorkflowableEntityType(ybo.EntityType, metaclass=workflowable_definition):
    """Use this base class instead of :class:`EntityType` to have workflow
    relations (i.e. `in_state`, `wf_info_for` and `custom_workflow`) on your
    entity type.
    """

    __abstract__ = True


def make_workflowable(cls, in_state_descr=None):
    """Adds workflow relations as :class:`WorkflowableEntityType`, but usable on
    existing classes which are not using that base class.
    """
    existing_rels = set(rdef.name for rdef in cls.__relations__)
    # let relation types defined in cw.schemas.workflow carrying
    # cardinality, constraints and other relation definition properties
    etype = getattr(cls, "name", cls.__name__)
    if "custom_workflow" not in existing_rels:
        rdef = ybo.RelationDefinition(etype, "custom_workflow", "Workflow")
        yams_add_relation(cls.__relations__, rdef)
    if "in_state" not in existing_rels:
        rdef = ybo.RelationDefinition(
            etype, "in_state", "State", description=in_state_descr
        )
        yams_add_relation(cls.__relations__, rdef)
    if "wf_info_for" not in existing_rels:
        rdef = ybo.RelationDefinition("TrInfo", "wf_info_for", etype)
        yams_add_relation(cls.__relations__, rdef)


# schema loading ##############################################################

CONSTRAINTS["RQLConstraint"] = RQLConstraint
CONSTRAINTS["RQLUniqueConstraint"] = RQLUniqueConstraint
CONSTRAINTS["RQLVocabularyConstraint"] = RQLVocabularyConstraint
# don't want MultipleStaticVocabularyConstraint in cw yams schema
CONSTRAINTS.pop("MultipleStaticVocabularyConstraint", None)
SchemaLoader.context.update(CONSTRAINTS)


class BootstrapSchemaLoader(SchemaLoader):
    """cubicweb specific schema loader, loading only schema necessary to read
    the persistent schema
    """

    schema_class = CubicWebSchema

    # retro-compatibility for yams <= 0.45.*
    # this situation is not handled by our deprecation tools
    schemacls = schema_class

    def load(
        self, config, modnames=(["cubicweb", "cubicweb.schemas.bootstrap"],), **kwargs
    ):
        """return a Schema instance from the schema definition read
        from <directory>
        """
        return super(BootstrapSchemaLoader, self).load(
            modnames, name=config.appid, register_base_types=False, **kwargs
        )

    def unhandled_file(self, filepath):
        """called when a file without handler associated has been found"""
        self.warning("ignoring file %r", filepath)

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


class CubicWebSchemaLoader(BootstrapSchemaLoader):
    """cubicweb specific schema loader, automatically adding metadata to the
    instance's schema
    """

    def load(self, config, **kwargs):
        """return a Schema instance from the schema definition read
        from <directory>
        """
        self.info("loading %s schemas", ", ".join(config.cubes()))
        try:
            return super(CubicWebSchemaLoader, self).load(
                config, config.schema_modnames(), **kwargs
            )
        finally:
            # we've to cleanup modules imported from cubicweb.schemas as well
            cleanup_sys_modules([join(cubicweb.CW_SOFTWARE_ROOT, "schemas")])

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


set_log_methods(CubicWebSchemaLoader, getLogger("cubicweb.schemaloader"))
set_log_methods(BootstrapSchemaLoader, getLogger("cubicweb.bootstrapschemaloader"))
set_log_methods(RQLExpression, getLogger("cubicweb.schema"))

# _() is just there to add messages to the catalog, don't care about actual
# translation
MAY_USE_TEMPLATE_FORMAT = set(("managers",))
NEED_PERM_FORMATS = [_("text/cubicweb-page-template")]


@monkeypatch(FormatConstraint)
def vocabulary(self, entity=None, form=None):
    cw = None
    if form is None and entity is not None:
        cw = entity._cw
    elif form is not None:
        cw = form._cw
    if cw is not None:
        if hasattr(cw, "write_security"):  # test it's a session and not a request
            # cw is a server session
            hasperm = (
                not cw.write_security
                or not cw.is_hook_category_activated("integrity")
                or cw.user.matching_groups(MAY_USE_TEMPLATE_FORMAT)
            )
        else:
            hasperm = cw.user.matching_groups(MAY_USE_TEMPLATE_FORMAT)
        if hasperm:
            return self.regular_formats + tuple(NEED_PERM_FORMATS)
    return self.regular_formats



from pkg_resources import VersionConflict

try:
    from cwtags import __pkginfo__
except ImportError:
    pass
else:
    if __pkginfo__.numversion < (1, 2, 3):
        raise VersionConflict(
            "This version ({}) of cwtags is incompatible with CubicWeb >= 3.32. "
            "Please, upgrade your version of cwtags to >= 1.2.3 before running "
            "this migration.".format(__pkginfo__.version)
        )



# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

"""This module imports the Diseasome data into a CubicWeb instance."""

# Python imports
import sys
import argparse

# Logilab import, for timing
from logilab.common.decorators import timed

# CubicWeb imports
import cubicweb.dataimport as cwdi
from cubicweb_dataio import dataimport as mcwdi

# Diseasome parser import
import diseasome_parser as parser


def _is_of_class(instance, class_name):
    """Helper function to determine whether an instance is
    of a specified class or not.
    Returns a True if this is the case and False otherwise.
    """
    if instance.__class__.__name__ == class_name:
        return True
    else:
        return False


@timed
def diseasome_import(session, file_name, store):
    """Main function for importing Diseasome data.

    It uses the Diseasome data parser to get the contents of the
    data from a file, then uses a store for importing the data
    into a CubicWeb instance.

    >>> diseasome_import(session, 'file_name', Store)

    """
    exturis = dict(session.execute("Any U, X WHERE X is ExternalUri, X uri U"))
    uri_to_eid = {}
    uri_to_etype = {}
    all_relations = {}
    etypes = {
        (
            "http://www4.wiwiss.fu-berlin.de/" "diseasome/resource/diseasome/genes"
        ): "Gene",
        (
            "http://www4.wiwiss.fu-berlin.de/" "diseasome/resource/diseasome/diseases"
        ): "Disease",
    }
    # Read the parsed data
    for entity, relations in parser.entities_from_rdf(file_name, ("gene", "disease")):
        uri = entity.get("cwuri", None)
        types = list(relations.get("types", []))
        if not types:
            continue
        etype = etypes.get(types[0])
        if not etype:
            sys.stderr.write("Entity type %s not recognized.", types[0])
            sys.stderr.flush()
        if _is_of_class(store, "MassiveObjectStore"):
            for relation in set(relations).intersection(
                (
                    "classes",
                    "possible_drugs",
                    "omim",
                    "omim_page",
                    "chromosomal_location",
                    "same_as",
                    "gene_id",
                    "hgnc_id",
                    "hgnc_page",
                )
            ):
                store.init_rtype_table(etype, relation, "ExternalUri")
            for relation in set(relations).intersection(("subtype_of",)):
                store.init_rtype_table(etype, relation, "Disease")
            for relation in set(relations).intersection(("associated_genes",)):
                store.init_rtype_table(etype, relation, "Gene")
        # Create the entities
        ent = store.create_entity(etype, **entity)
        if not _is_of_class(store, "MassiveObjectStore"):
            uri_to_eid[uri] = ent.eid
            uri_to_etype[uri] = ent.cw_etype
        else:
            uri_to_eid[uri] = uri
            uri_to_etype[uri] = etype
        # Store relations for after
        all_relations[uri] = relations
    # Perform a first commit, of the entities
    store.flush()
    kwargs = {}
    for uri, relations in all_relations.items():
        from_eid = uri_to_eid.get(uri)
        # ``subjtype`` should be initialized if ``SQLGenObjectStore`` is used
        # and there are inlined relations in the schema.
        # If ``subjtype`` is not given, while ``SQLGenObjectStore`` is used
        # and there are inlined relations in the schema, the store
        # tries to infer the type of the subject, but this does not always
        # work, e.g. when there are several object types for the relation.
        # ``subjtype`` is ignored for other stores, or if there are no
        # inlined relations in the schema.
        kwargs["subjtype"] = uri_to_etype.get(uri)
        if not from_eid:
            continue
        for rtype, rels in relations.items():
            if rtype in (
                "classes",
                "possible_drugs",
                "omim",
                "omim_page",
                "chromosomal_location",
                "same_as",
                "gene_id",
                "hgnc_id",
                "hgnc_page",
            ):
                for rel in list(rels):
                    if rel not in exturis:
                        # Create the "ExternalUri" entities, which are the
                        # objects of the relations
                        extu = store.create_entity("ExternalUri", uri=rel)
                        if not _is_of_class(store, "MassiveObjectStore"):
                            rel_eid = extu.eid
                        else:
                            # For the "MassiveObjectStore", the EIDs are
                            # in fact the URIs.
                            rel_eid = rel
                        exturis[rel] = rel_eid
                    else:
                        rel_eid = exturis[rel]
                    # Create the relations that have "ExternalUri"s as objects
                    if not _is_of_class(store, "MassiveObjectStore"):
                        store.relate(from_eid, rtype, rel_eid, **kwargs)
                    else:
                        store.relate_by_iid(from_eid, rtype, rel_eid)
            elif rtype in ("subtype_of", "associated_genes"):
                for rel in list(rels):
                    to_eid = uri_to_eid.get(rel)
                    if to_eid:
                        # Create relations that have objects of other type
                        # than "ExternalUri"
                        if not _is_of_class(store, "MassiveObjectStore"):
                            store.relate(from_eid, rtype, to_eid, **kwargs)
                        else:
                            store.relate_by_iid(from_eid, rtype, to_eid)
                    else:
                        sys.stderr.write(
                            "Missing entity with URI %s "
                            "for relation %s" % (rel, rtype)
                        )
                        sys.stderr.flush()
    # Perform a second commit, of the "ExternalUri" entities.
    # when the stores in the CubicWeb ``dataimport`` module are used,
    # relations are also committed.
    store.flush()
    # If the ``MassiveObjectStore`` is used, then entity and relation metadata
    # are pushed as well. By metadata we mean information on the creation
    # time and author.
    if _is_of_class(store, "MassiveObjectStore"):
        for relation in (
            "classes",
            "possible_drugs",
            "omim",
            "omim_page",
            "chromosomal_location",
            "same_as",
        ):
            # Afterwards, relations are actually created in the database.
            store.convert_relations("Disease", relation, "ExternalUri", "cwuri", "uri")
        store.convert_relations("Disease", "subtype_of", "Disease", "cwuri", "cwuri")
        store.convert_relations("Disease", "associated_genes", "Gene", "cwuri", "cwuri")
        for relation in ("gene_id", "hgnc_id", "hgnc_page", "same_as"):
            store.convert_relations("Gene", relation, "ExternalUri", "cwuri", "uri")
        # Clean up temporary tables in the database
        store.cleanup()


if __name__ == "__main__":
    # Change sys.argv so that ``cubicweb-ctl shell`` can work out the options
    # we give to our ``diseasome_import.py`` script.
    sys.argv = [arg for arg in sys.argv[sys.argv.index("--") - 1 :] if arg != "--"]
    PARSER = argparse.ArgumentParser(description="Import Diseasome data")
    PARSER.add_argument("-df", "--datafile", type=str, help="RDF data file name")
    PARSER.add_argument(
        "-st", "--store", type=str, default="RQLObjectStore", help="data import store"
    )
    ARGS = PARSER.parse_args()
    if ARGS.datafile:
        FILENAME = ARGS.datafile
        if ARGS.store in (st + "ObjectStore" for st in ("RQL", "NoHookRQL", "SQLGen")):
            IMPORT_STORE = getattr(cwdi, ARGS.store)(session)
        elif ARGS.store == "MassiveObjectStore":
            IMPORT_STORE = mcwdi.MassiveObjectStore(session)
        else:
            sys.exit("Import store unknown")
        diseasome_import(session, FILENAME, IMPORT_STORE)
    else:
        sys.exit("Data file not found or not specified")



from yams.constraints import UniqueConstraint

from cubicweb.schema import PURE_VIRTUAL_RTYPES
from cubicweb.server.checkintegrity import expected_indexes, database_indexes

source = repo.system_source

for rschema in schema.relations():
    if rschema.rule or rschema in PURE_VIRTUAL_RTYPES:
        continue
    if rschema.final or rschema.inlined:
        for rdef in rschema.rdefs.values():
            table = f"cw_{rdef.subject}"
            column = f"cw_{rdef.rtype}"
            if any(isinstance(cstr, UniqueConstraint) for cstr in rdef.constraints):
                source.create_index(cnx, table, column, unique=True)
                commit(ask_confirm=False)
            if rschema.inlined or rdef.indexed:
                source.create_index(cnx, table, column)
                commit(ask_confirm=False)

schema_indices = expected_indexes(cnx)
db_indices = database_indexes(cnx)
for additional_index in db_indices - set(schema_indices):
    try:
        sql(f"DROP INDEX {additional_index}")
        commit()
    except Exception:
        # ignore if this is not an index but a constraint
        pass

if source.dbhelper == "postgres" and "appears_words_idx" not in db_indices:
    sql("CREATE INDEX appears_words_idx ON appears USING gin(words)")
    db_indices.add("appears_words_idx")

for missing_index in set(schema_indices) - db_indices:
    print("WARNING: missing index", missing_index)



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Common utilies to format / send emails."""

from base64 import b64encode, b64decode
from email.header import Header
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate
from socket import gethostname
from time import time


def header(ustring):
    return Header(ustring, "utf-8")


def addrheader(addr, uname=None):
    # even if an email address should be ascii, encode it using utf8 since
    # automatic tests may generate non ascii email address
    if uname:
        val = f"{header(uname).encode()} <{addr}>"
    else:
        val = addr
    assert isinstance(val, str)  # bytes in py2, ascii-encoded unicode in py3
    return val


def construct_message_id(appid, eid, withtimestamp=True):
    if withtimestamp:
        addrpart = f"eid={eid}&timestamp={time():.10f}"
    else:
        addrpart = f"eid={eid}"
    # we don't want any equal sign nor trailing newlines
    leftpart = (
        b64encode(addrpart.encode("ascii"), b".-").decode("ascii").rstrip().rstrip("=")
    )
    return f"<{leftpart}@{appid}.{gethostname()}>"


def parse_message_id(msgid, appid):
    if msgid[0] == "<":
        msgid = msgid[1:]
    if msgid[-1] == ">":
        msgid = msgid[:-1]
    try:
        values, qualif = msgid.split("@")
        padding = len(values) % 4
        values = b64decode(str(values + "=" * padding), ".-").decode("ascii")
        values = dict(v.split("=") for v in values.split("&"))
        fromappid, host = qualif.split(".", 1)
    except Exception:
        return None
    if appid != fromappid or host != gethostname():
        return None
    return values


def format_mail(
    uinfo,
    to_addrs,
    content,
    subject="",
    cc_addrs=(),
    msgid=None,
    references=(),
    config=None,
):
    """Sends an Email to 'e_addr' with content 'content', and subject 'subject'

    to_addrs and cc_addrs are expected to be a list of email address without
    name
    """
    assert isinstance(content, str), repr(content)
    msg = MIMEText(content.encode("UTF-8"), "plain", "UTF-8")
    # safety: keep only the first newline
    try:
        subject = subject.splitlines()[0]
        msg["Subject"] = header(subject)
    except IndexError:
        pass  # no subject
    if uinfo.get("email"):
        email = uinfo["email"]
    elif config and config["sender-addr"]:
        email = config["sender-addr"]
    else:
        email = ""
    if uinfo.get("name"):
        name = uinfo["name"]
    elif config and config["sender-name"]:
        name = config["sender-name"]
    else:
        name = ""
    if config and config["sender-addr"] and config["sender-addr"] != email:
        msg["From"] = addrheader(config["sender-addr"], name)
    else:
        msg["From"] = addrheader(email, name)
    if config is not None and config["sender-x-cw-header"]:
        msg["X-CW"] = config["sender-x-cw-header"]
    unique_addrs = lambda addrs: sorted({addr for addr in addrs if addr is not None})
    msg["To"] = ", ".join(addrheader(addr) for addr in unique_addrs(to_addrs))
    if cc_addrs:
        msg["Cc"] = ", ".join(addrheader(addr) for addr in unique_addrs(cc_addrs))
    if msgid:
        msg["Message-id"] = msgid
    if references:
        msg["References"] = ", ".join(references)
    msg["Date"] = formatdate()
    return msg


class HtmlEmail(MIMEMultipart):
    def __init__(
        self,
        subject,
        textcontent,
        htmlcontent,
        sendermail=None,
        sendername=None,
        recipients=None,
        ccrecipients=None,
    ):
        MIMEMultipart.__init__(self, "related")
        self["Subject"] = header(subject)
        self.preamble = "This is a multi-part message in MIME format."
        # Attach alternative text message
        alternative = MIMEMultipart("alternative")
        self.attach(alternative)
        msgtext = MIMEText(textcontent.encode("UTF-8"), "plain", "UTF-8")
        alternative.attach(msgtext)
        # Attach html message
        msghtml = MIMEText(htmlcontent.encode("UTF-8"), "html", "UTF-8")
        alternative.attach(msghtml)
        if sendermail or sendername:
            self["From"] = addrheader(sendermail, sendername)
        if recipients:
            self["To"] = ", ".join(
                addrheader(addr) for addr in recipients if addr is not None
            )
        if ccrecipients:
            self["Cc"] = ", ".join(
                addrheader(addr) for addr in ccrecipients if addr is not None
            )

    def attach_image(self, data, htmlId):
        image = MIMEImage(data)
        image.add_header("Content-ID", f"<{htmlId}>")
        self.attach(image)



"""entity/relation schemas to store email in an cubicweb instance

:organization: Logilab
:copyright: 2006-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: https://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

# pylint: disable-msg=E0611,F0401
from yams.buildobjs import (
    SubjectRelation,
    RelationType,
    EntityType,
    String,
    Datetime,
    Int,
    RelationDefinition,
)
from yams.reader import context

from cubicweb import _
from cubicweb.schema import ERQLExpression


class Email(EntityType):
    """electronic mail"""

    subject = String(fulltextindexed=True)
    date = Datetime(description=_("UTC time on which the mail was sent"))
    messageid = String(required=True, indexed=True)
    headers = String(description=_("raw headers"))

    sender = SubjectRelation("EmailAddress", cardinality="?*")
    # an email with only Bcc is acceptable, don't require any recipients
    recipients = SubjectRelation("EmailAddress")
    cc = SubjectRelation("EmailAddress")

    parts = SubjectRelation("EmailPart", cardinality="*1", composite="subject")
    attachment = SubjectRelation("File")

    reply_to = SubjectRelation("Email", cardinality="?*")
    cites = SubjectRelation("Email")
    in_thread = SubjectRelation("EmailThread", cardinality="?*")


class EmailPart(EntityType):
    """an email attachment"""

    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),  # XXX if E parts X, U has_read_permission E
        "add": (
            "managers",
            ERQLExpression("E parts X, U has_update_permission E"),
        ),
        "delete": ("managers", ERQLExpression("E parts X, U has_update_permission E")),
        "update": (
            "managers",
            "owners",
        ),
    }

    content = String(fulltextindexed=True)
    content_format = String(required=True, maxsize=50)
    ordernum = Int(required=True)
    alternative = SubjectRelation("EmailPart", symmetric=True)


class EmailThread(EntityType):
    """discussion thread"""

    title = String(required=True, indexed=True, fulltextindexed=True)
    see_also = SubjectRelation("EmailThread")
    forked_from = SubjectRelation("EmailThread", cardinality="?*")


class parts(RelationType):
    """ """

    fulltext_container = "subject"


class sender(RelationType):
    """ """

    inlined = True


class in_thread(RelationType):
    """ """

    inlined = True


class reply_to(RelationType):
    """ """

    inlined = True


class generated_by(RelationType):
    """mark an entity as generated from an email"""

    cardinality = "?*"
    subject = ("TrInfo",)
    object = "Email"


# if comment is installed
if "Comment" in context.defined:

    class comment_generated_by(RelationDefinition):
        subject = "Comment"
        name = "generated_by"
        object = "Email"
